from __future__ import annotations

import typing as t
from contextlib import ContextDecorator
from multiprocessing import Process
from pathlib import Path

from pxblat.extc import ServerOption, UsageStats, pystartServer

from .basic import (
    check_port_in_use,
    files,
    find_free_port,
    server_query,
    status_server,
    stop_server,
    wait_server_ready,
)

if t.TYPE_CHECKING:
    from .status import Status


def _pystartServer(
    hostName: str,
    portName: str,
    seqFiles: list[str],
    options: ServerOption,
    stats: UsageStats,
):
    pystartServer(hostName, portName, len(seqFiles), seqFiles, options, stats)


def create_server_option() -> ServerOption:
    """Creates a new ServerOption object with default values.

    Returns:
        ServerOption: A new ServerOption object with default values.

    See Also:
        :class:`.ServerOption`
    """
    return ServerOption()


class Server(ContextDecorator):
    """A context manager and decorator for managing a server process.

    This class can be used as a context manager or decorator to manage a server process. It starts the server with the given
    options, and can run it as a daemon process or block until it is ready.

    Attributes:
        host (str): The hostname or IP address to bind the server to.
        port (int): The port number to bind the server to.
        two_bit (Path | str): The path to the 2bit file or the URL of the 2bit file.
        option (ServerOption): The options to use when starting the server.
        daemon (bool, optional): Whether to run the server as a daemon process. Defaults to True.
        use_others (bool, optional): Whether to allow other users to access the server. Defaults to False.
        timeout (int, optional): The number of seconds to wait for the server to start. Defaults to 60.
        block (bool, optional): Whether to block until the server is ready. Defaults to False.

    Raises:
        ValueError: If the given two_bit file or URL is invalid.
        OSError: If there is an error starting the server process.

    Order:
        -10
    """

    def __init__(
        self,
        host: str,
        port: int,
        two_bit: Path | str,
        *,
        can_stop: bool = True,
        mask: bool = False,
        tile_size: int = 11,
        step_size: int = 11,
        max_aa_size: int = 8000,
        max_dna_hits: int = 100,
        max_gap: int = 2,
        max_nt_size: int = 40000,
        max_trans_hits: int = 200,
        min_match: int = 2,
        rep_match: int = 0,
        seq_log: bool = False,
        ip_log: bool = False,
        debug_log: bool = False,
        trans: bool = False,
        syslog: bool = False,
        no_simp_rep_mask: bool = False,
        log: str | Path | None = None,
        log_facility: str | None = None,
        per_seq_max: str | Path | None = None,
        index_file: str | Path | None = None,
        daemon=True,
        use_others: bool = False,
        timeout: int = 60,
        block: bool = False,
    ) -> None:
        """Initializes a gfServer object with the given parameters.

        Args:
            host (str): The hostname or IP address to bind the server to.
            port (int): The port number to bind the server to.
            two_bit (Path | str): The path to the 2bit file or the URL of the 2bit file.
            can_stop (bool, optional): Whether to allow the server to be stopped. Defaults to True.
            mask (bool, optional): Whether to use masking from the 2bit file. Defaults to False.
            tile_size (int, optional): The size of n-mers to index. Defaults to 11 for nucleotides, 4 for proteins (or translated nucleotides).
            step_size (int, optional): The spacing between tiles. Defaults to tileSize.
            max_aa_size (int, optional): The maximum size of protein or translated DNA queries. Defaults to 8000.
            max_dna_hits (int, optional): The maximum number of hits for a DNA query that are sent from the server. Defaults to 100.
            max_gap (int, optional): The number of insertions or deletions allowed between n-mers. Defaults to 2 for nucleotides, 0 for proteins.
            max_nt_size (int, optional): The maximum size of untranslated DNA query sequence. Defaults to 40000.
            max_trans_hits (int, optional): The maximum number of hits for a translated query that are sent from the server. Defaults to 200.
            min_match (int, optional): The number of n-mer matches that trigger detailed alignment. Defaults to 2 for nucleotides, 3 for proteins.
            rep_match (int, optional): The number of occurrences of a tile (n-mer) that triggers repeat masking the tile. Defaults to 0.
            seq_log (bool, optional): Whether to include sequences in the log file (not logged with syslog). Defaults to False.
            ip_log (bool, optional): Whether to include user's IP in the log file (not logged with syslog). Defaults to False.
            debug_log (bool, optional): Whether to include debugging info in the log file. Defaults to False.
            trans (bool, optional): Whether to translate database to protein in 6 frames, and it is best to run this on RepeatMasked data. Defaults to False.
            syslog (bool, optional): Whether to log to syslog. Defaults to False.
            no_simp_rep_mask (bool, optional): Whether to suppress simple repeat masking. Defaults to False.
            log (str | Path | None, optional): The path to the log file that records server requests. Defaults to None.
            log_facility (str | None, optional): The syslog facility to log to. Defaults to None.
            per_seq_max (str | Path | None, optional): The path to a file that contains one seq filename (possibly with ':seq' suffix) per line. Defaults to None.
            index_file (str | Path | None, optional): The path to the index file created by `gfServer index`.
                Saving index can speed up `gfServer` startup by two orders of magnitude. Defaults to None.
            daemon (bool, optional): Whether to run the server as a daemon process. Defaults to True.
            use_others (bool, optional): Whether to allow other users to access the server. Defaults to False.
            timeout (int, optional): The number of seconds to wait for the server to start. Defaults to 60.
            block (bool, optional): Whether to block until the server is ready. Defaults to False.

        Raises:
            ValueError: If the given two_bit file or URL is invalid.
            OSError: If there is an error starting the server process.

        Returns:
            None

        Examples:
            Create a server object with options.

            >>> from pxblat import Server
            >>> server = Server("localhost", 65000, "tests/data/test_ref.2bit", can_stop=True, step_size=5)
            >>> server.start()
            >>> server.wait_ready()
            >>> server.stop()
            >>> server.can_stop
            True
            >>> server.step_size = 10
            >>> server.step_size
            10
        """
        self._host = host
        self._port = port

        self.two_bit = two_bit

        log = "" if log is None else str(log)
        log_facility = "" if log_facility is None else str(log_facility)
        per_seq_max = "" if per_seq_max is None else str(per_seq_max)
        index_file = "" if index_file is None else str(index_file)

        self.option = (
            create_server_option()
            .withCanStop(can_stop)
            .withLog(log)
            .withLogFacility(log_facility)
            .withMask(mask)
            .withMaxAaSize(max_aa_size)
            .withMaxDnaHits(max_dna_hits)
            .withMaxGap(max_gap)
            .withMaxNtSize(max_nt_size)
            .withMaxTransHits(max_trans_hits)
            .withMinMatch(min_match)
            .withRepMatch(rep_match)
            .withSeqLog(seq_log)
            .withIpLog(ip_log)
            .withDebugLog(debug_log)
            .withTileSize(tile_size)
            .withStepSize(step_size)
            .withTrans(trans)
            .withSyslog(syslog)
            .withPerSeqMax(per_seq_max)
            .withNoSimpRepMask(no_simp_rep_mask)
            .withIndexFile(index_file)
        )

        self.stat = UsageStats()

        self.use_others = use_others
        self.timeout = timeout
        self.daemon = daemon

        self._block = block
        self._is_ready = False
        self._is_open = True
        self._process = None

    @property
    def host(self):
        """The hostname or IP address to bind the server to."""
        return self._host

    @host.setter
    def host(self, value: str):
        """Sets the hostname or IP address to bind the server to."""
        self._host = value

    @property
    def port(self) -> int:
        """The port number to bind the server to."""
        return self._port

    @port.setter
    def port(self, value: int):
        """Sets the port number to bind the server to."""
        self._port = value

    def _start_b(self):
        """Start server in blocking mode."""
        two_bit_file = self.two_bit if isinstance(self.two_bit, str) else self.two_bit.as_posix()
        try:
            if check_port_in_use(self.host, self.port):
                if self.use_others:
                    self._is_open = False
                    # WARN: Use server that is already open. However, the server may be not opened by gfServer <05-16-23>
                    # Hence, the `wait_server_ready` may be timeout.

                else:
                    self._is_open = True
                    new_port = find_free_port(self.host, start=self.port + 1)
                    self.port = new_port
                    pystartServer(
                        self.host,
                        str(self.port),
                        1,
                        [two_bit_file],
                        self.option,
                        self.stat,
                    )
            else:
                pystartServer(
                    self.host,
                    str(self.port),
                    1,
                    [two_bit_file],
                    self.option,
                    self.stat,
                )
        except Exception as e:
            raise e

    def _start_nb(self):
        two_bit_file = self.two_bit if isinstance(self.two_bit, str) else self.two_bit.as_posix()
        try:
            if check_port_in_use(self.host, self.port):
                if self.use_others:
                    self._is_open = False
                else:
                    self._is_open = True
                    new_port = find_free_port(self._host, start=self.port + 1)
                    self.port = new_port
                    host = self.host
                    port = self.port

                    self._process = Process(
                        target=_pystartServer,
                        args=(
                            host,
                            str(port),
                            [two_bit_file],
                            self.option,
                            self.stat,
                        ),
                        daemon=self.daemon,
                    )

            else:
                self._is_open = True
                self._process = Process(
                    target=_pystartServer,
                    args=(
                        self.host,
                        str(self.port),
                        [two_bit_file],
                        self.option,
                        self.stat,
                    ),
                    daemon=self.daemon,
                )
        except Exception as e:
            raise e

        else:
            if self._process is not None:
                self._process.start()

    def _check(self):
        if not Path(self.two_bit).exists():
            msg = f"Invalid two_bit file: {self.two_bit}"
            raise FileNotFoundError(msg)

    def start(self):
        """Starts the gfServer instance in either blocking or non-blocking mode.

        If the server is set to non-blocking mode, it will start the server in a separate process.
        If the server is set to blocking mode, it will start the server in the current process.

        Raises:
            ValueError: If the given two_bit file or URL is invalid.
        """
        self.option.build()
        if not self._block:
            self._start_nb()
        else:
            self._start_b()

    def stop(self):
        """Stops the gfServer instance if it is running.

        This method sends a stop signal to the server process, causing it to terminate gracefully.

        See Also:
            :func:`stop_server` is a free function to stop a server.
        """
        if self._is_open:
            stop_server(self.host, self.port)

        if self._process is not None:
            self._process.terminate()

        self._is_open = False
        self._is_ready = False

    def status(self, *, instance=False) -> dict[str, str] | Status:
        """Retrieves the status of the gfServer instance.

        Args:
            instance (bool, optional): If True, returns a Status object. If False, returns a dictionary with status information. Defaults to False.

        Returns:
            t.Union[t.Dict[str, str], Status]: The status of the gfServer instance, either as a dictionary or a Status object.

        See Also:
            :func:`status_server` is a free function to query server status.
        """
        return status_server(self.host, self.port, self.option, instance=instance)

    def files(self) -> list[str]:
        """Retrieves the list of files served by the gfServer instance.

        Returns:
            list[str]: A list of file names served by the gfServer instance.

        See Also:
            :func:`files` is a free function to query file status for server.
        """
        return files(self.host, self.port)

    def query(self, intype: str, faName: str, *, isComplex: bool, isProt: bool) -> str:
        """Queries the gfServer instance with the given parameters.

        Args:
            intype (str): The type of input sequence. Must be one of "dna", "rna", or "protein".
            faName (str): The name of the input sequence.
            isComplex (bool): Whether the input sequence is complex.
            isProt (bool): Whether the input sequence is a protein sequence.

        Returns:
            str: The result of the query as a string.
        """
        return server_query(
            intype,
            self.host,
            self.port,
            faName,
            isComplex=isComplex,
            isProt=isProt,
        )

    def is_ready(self) -> bool:
        """Returns True if the server is ready to accept queries, False otherwise.

        Returns:
            bool: True if the server is ready to accept queries, False otherwise.
        """
        return self._is_ready

    def is_open(self) -> bool:
        """Returns True if the server is open, False otherwise.

        Returns:
            bool: True if the server is open, False otherwise.
        """
        return self._is_open

    def wait_ready(self, timeout: int = 60, *, restart: bool = False):
        """Wait server ready in block mode.

        Args:
            timeout: Timeout for wait server ready.
            restart: If timeout, restart server and wait again.

        Raises:
            RuntimeError: If server is not opened by gfServer or the server takes longer time to be ready, the `wait_server_ready` may be timeout.
            If `restart` is True, the server will be restarted and wait again.
        """
        if not self._is_ready:
            try:
                wait_server_ready(self.host, self.port, timeout, self.option)
            except RuntimeError as e:
                if restart and self.use_others:
                    self.use_others = False
                    self.start()
                    self.wait_ready(timeout * 2, restart=restart)
                else:
                    msg = f"Timeout for Waiting for {self.host} {self.port} server ready due to server is not opened by gfServer or need longer time to wait"
                    raise RuntimeError(
                        msg,
                    ) from e
            else:
                self._is_ready = True

    def __str__(self) -> str:
        """Return server option as a string."""
        return f"Server({self.host}, {self.port}, ready: {self.is_ready()} open: {self.is_open()}\n{self.option})"

    __repr__ = __str__

    def __enter__(self):
        """Starts the gfServer instance in blocking mode when used as a context manager."""
        self.start()
        return self

    def __exit__(self, *exc):
        """Stops the gfServer."""
        self.stop()

    # fmt: off
    @property
    def can_stop(self) -> bool: return self.option.canStop
    @can_stop.setter
    def can_stop(self, value: bool): self.option.canStop = value
    @property
    def log(self) -> str: return self.option.log
    @log.setter
    def log(self, value: str): self.option.log = value
    @property
    def log_facility(self) -> str: return self.option.logFacility
    @log_facility.setter
    def log_facility(self, value: str): self.option.logFacility = value
    @property
    def mask(self) -> bool: return self.option.mask
    @mask.setter
    def mask(self, value: bool): self.option.mask = value
    @property
    def max_aa_size(self) -> int: return self.option.maxAaSize
    @max_aa_size.setter
    def max_aa_size(self, value: int): self.option.maxAaSize = value
    @property
    def max_dna_hits(self) -> int: return self.option.maxDnaHits
    @max_dna_hits.setter
    def max_dna_hits(self, value: int): self.option.maxDnaHits = value
    @property
    def max_gap(self) -> int: return self.option.maxGap
    @max_gap.setter
    def max_gap(self, value: int): self.option.maxGap = value
    @property
    def max_nt_size(self) -> int: return self.option.maxNtSize
    @max_nt_size.setter
    def max_nt_size(self, value: int): self.option.maxNtSize = value
    @property
    def max_trans_hits(self) -> int: return self.option.maxTransHits
    @max_trans_hits.setter
    def max_trans_hits(self, value: int): self.option.maxTransHits = value
    @property
    def min_match(self) -> int: return self.option.minMatch
    @min_match.setter
    def min_match(self, value: int): self.option.minMatch = value
    @property
    def rep_match(self) -> int: return self.option.repMatch
    @rep_match.setter
    def rep_match(self, value: int): self.option.repMatch = value
    @property
    def seq_log(self) -> bool: return self.option.seqLog
    @seq_log.setter
    def seq_log(self, value: bool): self.option.seqLog = value
    @property
    def ip_log(self) -> bool: return self.option.ipLog
    @ip_log.setter
    def ip_log(self, value: bool): self.option.ipLog = value
    @property
    def debug_log(self) -> bool: return self.option.debugLog
    @debug_log.setter
    def debug_log(self, value: bool): self.option.debugLog = value
    @property
    def trans(self) -> bool: return self.option.trans
    @trans.setter
    def trans(self, value: bool): self.option.trans = value
    @property
    def syslog(self) -> bool: return self.option.syslog
    @syslog.setter
    def syslog(self, value: bool): self.option.syslog= value
    @property
    def no_simp_rep_mask(self) -> bool: return self.option.noSimpRepMask
    @no_simp_rep_mask.setter
    def no_simp_rep_mask(self, value: bool): self.option.noSimpRepMask= value
    @property
    def per_seq_max(self) -> str: return self.option.perSeqMax
    @per_seq_max.setter
    def per_seq_max(self, value: str): self.option.perSeqMax = value
    @property
    def index_file(self) -> str: return self.option.indexFile
    @index_file.setter
    def index_file(self, value: str): self.option.indexFile= value
    # fmt: on



import os
import shlex
import sys
import typing
from contextlib import contextmanager
from ctypes.util import find_library
from functools import wraps
from pathlib import Path

import distutils
import setuptools
from pybind11.setup_helpers import auto_cpp_level
from pybind11.setup_helpers import ParallelCompile
from pybind11.setup_helpers import Pybind11Extension
from setuptools import Distribution
from setuptools import Extension
from setuptools.command.build_ext import build_ext as _build_ext


DEBUG = False


class PxblatExtensionBuilder(_build_ext):
    def build_extension(self, extension: setuptools.extension.Extension) -> None:  # type: ignore
        extension.library_dirs.append(self.build_lib)  # type: ignore
        super().build_extension(extension)

    def build_extensions(self) -> None:
        """
        Build extensions, injecting C++ std for Pybind11Extension if needed.
        """

        for ext in self.extensions:
            if hasattr(ext, "_cxx_level") and ext._cxx_level == 0:
                ext.cxx_std = auto_cpp_level(self.compiler)

        super().build_extensions()


def _get_pxblat_libname():
    builder = setuptools.command.build_ext.build_ext(Distribution())  # type: ignore
    full_name = builder.get_ext_filename("libpxblat")
    without_lib = full_name.split("lib", 1)[-1]
    without_so = without_lib.rsplit(".so", 1)[0]
    return without_so


def remove_env(key: str):
    """Remove environment variable."""
    env_cflags = os.environ.get("CFLAGS", "")
    env_cppflags = os.environ.get("CPPFLAGS", "")
    flags = shlex.split(env_cflags) + shlex.split(env_cppflags)

    for flag in flags:
        if flag.startswith(key):
            raise RuntimeError(f"Please remove {key} from CFLAGS and CPPFLAGS.")


@contextmanager
def change_dir(path: str):
    """Change directory."""
    save_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(save_dir)


def change_env(key: str, value: str):
    """Change environment variable."""

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_env = os.environ.get(key, None)
            os.environ[key] = old_env + " " + value if old_env else value
            func(*args, **kwargs)
            os.environ[key] = old_env if old_env else " "

        return wrapper

    return decorator


def get_files_by_suffix(
    path: typing.Union[Path, str], suffix: typing.List[str]
) -> typing.Iterator[str]:
    """Get bindings."""
    if isinstance(path, str):
        path = Path(path)

    for file in path.iterdir():
        if file.is_dir():
            yield from get_files_by_suffix(file, suffix)
        if file.suffix in suffix:
            yield file.as_posix()


def filter_files(files, exclude=None):
    if exclude is None:
        exclude = []

    for file in files:
        file_name = Path(file).name
        if file_name not in exclude:
            yield file


# Optional multithreaded build
def get_thread_count():
    try:
        import multiprocessing

        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass
    return 1


def _get_cxx_compiler():
    cc = distutils.ccompiler.new_compiler()  # type: ignore
    distutils.sysconfig.customize_compiler(cc)  # type: ignore
    return cc.compiler_cxx[0]  # type: ignore


def find_lib_in_conda(lib_name: str):
    conda_prefix = os.environ.get("CONDA_PREFIX", None)
    if conda_prefix is not None:
        conda_lib_dir = Path(conda_prefix) / "lib"

        if (conda_lib_dir / f"lib{lib_name}.a").exists():
            return conda_lib_dir

        if (conda_lib_dir / f"lib{lib_name}.so").exists():
            return conda_lib_dir

        if (conda_lib_dir / f"lib{lib_name}.dylib").exists():
            return conda_lib_dir

    return None


def find_available_library(lib_name: str):
    lib_path = find_library(lib_name)

    if lib_path is None:
        lib_path = find_lib_in_conda(lib_name)

    print(f"{lib_name} lib_path: {lib_path}")

    if not lib_path:
        raise RuntimeError(f"Cannot find {lib_name} library.")

    header_path = Path(lib_path).parent.parent / "include"

    return Path(lib_path).parent, header_path


def _extra_compile_args_for_libpxblat():
    return [
        "-D_FILE_OFFSET_BITS=64",
        "-D_LARGEFILE_SOURCE",
        "-D_GNU_SOURCE",
        "-DMACHTYPE_$(MACHTYPE)",
        "-DPXBLATLIB",
    ]


def _include_dirs_for_libpxblat():
    return [
        "src/pxblat/extc/include/core",
        "src/pxblat/extc/include/aux",
        "src/pxblat/extc/include/net",
    ]


def _include_dirs_for_pxblat():
    return [
        "src/pxblat/extc/bindings",
    ]


def _extra_compile_args_for_pxblat():
    flag = []
    if not DEBUG:
        flag.append("-DDBG_MACRO_DISABLE")
    return flag


ParallelCompile(f"{get_thread_count()}").install()

extra_compile_args = ["-pthread"]
hidden_visibility_args = []
include_dirs: list[str] = []
library_dirs: list[str] = []
python_module_link_args = []
base_library_link_args: list[str] = []
external_libraries = [
    "z",
    "hts",
    "ssl",
    "crypto",
    "m",
]

for lib in external_libraries:
    lib_library_dir, lib_include_dir = find_available_library(lib)
    library_dirs.append(lib_library_dir.as_posix())
    include_dirs.append(lib_include_dir.as_posix())

if sys.platform == "win32":
    raise RuntimeError("Windows is not supported.")
elif sys.platform == "darwin":
    # See https://conda-forge.org/docs/maintainer/knowledge_base.html#newer-c-features-with-old-sdk
    extra_compile_args.append("-D_LIBCPP_DISABLE_AVAILABILITY")
    extra_compile_args.append("-undefined dynamic_lookup")
    hidden_visibility_args.append("-fvisibility=hidden")
    config_vars = distutils.sysconfig.get_config_vars()  # type: ignore
    config_vars["LDSHARED"] = config_vars["LDSHARED"].replace("-bundle", "")  # type: ignore
    python_module_link_args.append("-bundle")
    builder = setuptools.command.build_ext.build_ext(Distribution())  # type: ignore
    full_name = builder.get_ext_filename("libpxblat")
    print(f"full_name: {full_name}")
    base_library_link_args.append(
        f"-Wl,-dylib_install_name,@loader_path/../{full_name}"
    )
    base_library_link_args.append("-dynamiclib")
else:
    hidden_visibility_args.append("-fvisibility=hidden")
    python_module_link_args.append("-Wl,-rpath,$ORIGIN/..")


def get_extension_modules():
    extension_modules = []

    """
    Extension module which is actually a plain C++ library without Python bindings
    """
    libpxblat_sources = (
        list(filter_files(get_files_by_suffix("src/pxblat/extc/src/core", [".c"])))
        + list(
            filter_files(
                get_files_by_suffix("src/pxblat/extc/src/aux", [".c"]),
                exclude=["net.c"],
            )
        )
        + list(filter_files(get_files_by_suffix("src/pxblat/extc/src/net", [".c"])))
    )

    pxblat_library = Extension(
        "libpxblat",
        language="c",
        sources=libpxblat_sources,
        include_dirs=include_dirs + _include_dirs_for_libpxblat(),
        extra_compile_args=_extra_compile_args_for_libpxblat() + extra_compile_args,
        extra_link_args=base_library_link_args,
        libraries=external_libraries,
        library_dirs=library_dirs,
    )

    pxblat_libs = [_get_pxblat_libname()]
    extension_modules.append(pxblat_library)

    """
    An extension module which contains the main Python bindings for libblat
    """
    pxblat_python_sources = [
        "src/pxblat/extc/bindings/faToTwoBit.cpp",
        "src/pxblat/extc/bindings/twoBitToFa.cpp",
        "src/pxblat/extc/bindings/gfServer.cpp",
        "src/pxblat/extc/bindings/pygfServer.cpp",
        "src/pxblat/extc/bindings/gfClient.cpp",
    ] + list(
        filter_files(get_files_by_suffix("src/pxblat/extc/bindings/binder", [".cpp"]))
    )

    pxblat_python = Pybind11Extension(
        "pxblat._extc",
        language="c++",
        sources=pxblat_python_sources,
        include_dirs=include_dirs
        + _include_dirs_for_libpxblat()
        + _include_dirs_for_pxblat(),
        extra_compile_args=extra_compile_args
        + hidden_visibility_args
        + _extra_compile_args_for_pxblat(),
        libraries=external_libraries + pxblat_libs,
        extra_link_args=python_module_link_args,
        library_dirs=library_dirs,
    )

    extension_modules.append(pxblat_python)
    return extension_modules


def build(setup_kwargs):
    """Build cpp extension."""
    ext_modules = get_extension_modules()
    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmdclass": {"build_ext": PxblatExtensionBuilder},
            "zip_safe": False,
            "package_data": {"pxblat": ["py.typed", "*so"]},
        }
    )



"""Toolkit functions for working with BLAT."""
from __future__ import annotations

from pathlib import Path

from pxblat.extc import TwoBitToFaOption, faToTwoBit, twoBitToFa

#     "Args",
#     "Arguments",
#     "Attention",
#     "Attributes",
#     "Caution",
#     "Danger",
#     "Error",
#     "Example",
#     "Examples",
#     "Hint",
#     "Important",
#     "Keyword Args",
#     "Keyword Arguments",
#     "Methods",
#     "Note",
#     "Notes",
#     "Return",
#     "Returns",
#     "Raises",
#     "References",
#     "See Also",
#     "Tip",
#     "Todo",
#     "Warning",
#     "Warnings",
#     "Warns",
#     "Yield",
#     "Yields",

__all__ = ["fa_to_two_bit", "two_bit_to_fa"]


def fa_to_two_bit(
    inFiles: list[str],
    outFile: str,
    *,
    noMask: bool = False,
    stripVersion: bool = False,
    ignoreDups: bool = False,
    useLong: bool = False,
):
    """Convert one or more FASTA files to two-bit format.

    Args:
        inFiles (List[str]): A list of paths to the input FASTA files.
        outFile (str): The path to the output two-bit file.
        noMask (bool, optional): If True, do not mask the output sequence. Defaults to False.
        stripVersion (bool, optional): If True, strip the version number from the sequence IDs. Defaults to False.
        ignoreDups (bool, optional): If True, ignore duplicate sequences in the input files. Defaults to False.
        useLong (bool, optional): If True, use the long format for the two-bit file. Defaults to False.

    Returns:
        None

    This function converts one or more input FASTA files to two-bit format and saves the result to the specified output file.
    The function takes several optional arguments that control the behavior of the conversion process, such as whether to mask
    the output sequence or strip the version number from the sequence IDs. The function returns None.

    Example:
        >>> fa_to_two_bit(['input.fasta'], 'output.2bit')
    """
    return faToTwoBit(inFiles, outFile, noMask, stripVersion, ignoreDups, useLong)


def two_bit_to_fa(
    inName: str | Path,
    outName: str | Path,
    option: TwoBitToFaOption,
):
    """Converts a .2bit file to a .fa file.

    This function takes the name of an input .2bit file and an output .fa file, and an option object,
    converts the .2bit file to .fa format using the `twoBitToFa` function, and saves the result to the output file.

    Args:
        inName (Union[str, Path]): The name or Path of the input .2bit file.
        outName (Union[str, Path]): The name or Path of the output .fa file.
        option (TwoBitToFaOption): The options to use when converting the .2bit file.

    Returns:
        The return value from the `twoBitToFa` function.
    """
    if isinstance(inName, Path):
        inName = str(inName)

    if isinstance(outName, Path):
        outName = str(outName)

    return twoBitToFa(inName, outName, option)



import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET",
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler()],
)

logger = logging.getLogger("pxblat")



from pathlib import Path

import typer

from pxblat import TwoBitToFaOption, two_bit_to_fa

# twoBitToFa - Convert all or part of .2bit file to fasta
# usage:
#    twoBitToFa input.2bit output.fa
# options:
#    -seq=name       Restrict this to just one sequence.
#    -start=X        Start at given position in sequence (zero-based).
#    -end=X          End at given position in sequence (non-inclusive).
#    -seqList=file   File containing list of the desired sequence names
#                    in the format seqSpec[:start-end], e.g. chr1 or chr1:0-189
#                    where coordinates are half-open zero-based, i.e. [start,end).
#    -noMask         Convert sequence to all upper case.
#    -bpt=index.bpt  Use bpt index instead of built-in one.
#    -bed=input.bed  Grab sequences specified by input.bed. Will exclude introns.
#    -bedPos         With -bed, use chrom:start-end as the fasta ID in output.fa.
#    -udcDir=/dir/to/cache  Place to put cache for remote bigBed/bigWigs.
# Sequence and range may also be specified as part of the input
# file name using the syntax:
#       /path/input.2bit:name
#    or
#       /path/input.2bit:name:start-end


# out: str = typer.Option(
#        default_option.outputFormat,
#        "--out",
#    ),


def twoBitToFa(
    input2bit: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        readable=True,
        metavar="input.2bit",
        help="The input 2bit file",
    ),
    outputfa: Path = typer.Argument(
        ...,
        dir_okay=False,
        writable=True,
        show_default=False,
        metavar="out.fa",
        help="The output fasta file",
    ),
    seq: str = typer.Option("", "--seq", help="Restrict this to just one sequence."),
    start: int = typer.Option(
        0,
        "--start",
        help="Start at given position in sequence (zero-based).",
    ),
    end: int = typer.Option(
        0,
        "--end",
        help="End at given position in sequence (non-inclusive).",
    ),
    seqList: str = typer.Option(
        "",
        "--seqList",
        help="File containing list of the desired sequence names",
    ),
    noMask: bool = typer.Option(
        False,
        "--noMask",
        help="Convert sequence to all upper case.",
    ),
    bpt: str = typer.Option("", "--bpt", help="Use bpt index instead of built-in one."),
    bed: str = typer.Option("", "--bed", help="Grab sequences specified by input.bed."),
    bedPos: bool = typer.Option(
        False,
        "--bedPos",
        help="With -bed, use chrom:start-end as the fasta ID in output.fa.",
    ),
    udcDir: str = typer.Option(
        "",
        "--udcDir",
        help="Place to put cache for remote bigBed/bigWigs.",
    ),
):
    """Convert all or part of .2bit file to fasta."""
    if seqList and not Path(seqList).exists():
        msg = f"{seqList} does not exist"
        raise typer.BadParameter(msg)

    if bpt and not Path(bpt).exists():
        msg = f"{bpt} does not exist"
        raise typer.BadParameter(msg)

    if bed and not Path(bed).exists():
        msg = f"{bed} does not exist"
        raise typer.BadParameter(msg)

    if udcDir and not Path(udcDir).exists():
        msg = f"{udcDir} does not exist"
        raise typer.BadParameter(msg)

    option = TwoBitToFaOption()
    option.withSeq(seq).withStart(start).withEnd(end).withSeqList(seqList).withNoMask(
        noMask,
    ).withBpt(bpt).withBed(bed).withBedPos(bedPos).withUdcDir(udcDir).build()

    two_bit_to_fa(input2bit, outputfa, option=option)



from dataclasses import dataclass, field

from mashumaro import (
    DataClassDictMixin,  # type: ignore
    field_options,  # type: ignore
)
from mashumaro.mixins.json import DataClassJSONMixin  # type: ignore


@dataclass
class Status(DataClassJSONMixin, DataClassDictMixin):
    """A data class representing the status of a server.

    Attributes:
        version (str): The version of the server.
        serverType (str): The type of the server.
        types (str): The type of the server, an alias for 'serverType'.
        host (str): The hostname or IP address of the server.
        port (int): The port number the server is listening on.
        tileSize (int): The tile size used by the server.
        stepSize (int): The step size used by the server.
        minMatch (int): The minimum match used by the server.
        pcr_requests (int): The number of PCR requests the server has handled, an alias for 'pcr requests'.
        blat_requests (int): The number of BLAT requests the server has handled, an alias for 'blat requests'.
        bases (int): The number of bases processed by the server.
        misses (int): The number of misses by the server.
        noSig (int): The number of 'noSig' (no signature) events by the server.
        trimmed (int): The number of trimmed events by the server.
        warnings (int): The number of warnings issued by the server.
    """

    version: str
    serverType: str
    types: str = field(metadata=field_options(alias="type"))
    host: str
    port: int
    tileSize: int
    stepSize: int
    minMatch: int
    pcr_requests: int = field(metadata=field_options(alias="pcr requests"))
    blat_requests: int = field(metadata=field_options(alias="blat requests"))
    bases: int
    misses: int
    noSig: int
    trimmed: int
    warnings: int



"""Bindings for ::cppbinding namespace"""
from __future__ import annotations
import pxblat._extc.cppbinding
import typing

__all__ = [
    "IntStruct",
    "TwoBitToFaOption",
    "UsageStats",
    "buildIndex",
    "faToTwoBit",
    "genoFindDirect",
    "genoPcrDirect",
    "getFileList",
    "getPortIx",
    "ClientOption",
    "gfServer",
    "ServerOption",
    "pcrServer",
    "pygetFileList",
    "pygfClient",
    "pygfClient2",
    "pygfClient_no_gil",
    "pyqueryServer",
    "pystartServer",
    "pystartServer_no_gil",
    "pystatusServer",
    "queryServer",
    "startServer",
    "statusServer",
    "stopServer",
    "test",
    "test_add",
    "test_exception",
    "test_no_gil",
    "test_stat",
    "test_stdout",
    "test_with_gil",
    "twoBitToFa",
]

class IntStruct:
    def __init__(self, arg0: int) -> None: ...
    pass

class TwoBitToFaOption:
    def __getstate__(self) -> tuple: ...
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, arg0: TwoBitToFaOption) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def __str__(self) -> str: ...
    def build(self) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::build() --> struct cppbinding::TwoBitToFaOption &
        """
    def to_string(self) -> str:
        """
        C++: cppbinding::TwoBitToFaOption::to_string() --> std::string
        """
    def withBed(self, bed: str) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withBed(std::string const &) --> struct cppbinding::TwoBitToFaOption &
        """
    def withBedPos(self, bedPos: bool) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withBedPos(bool) --> struct cppbinding::TwoBitToFaOption &
        """
    def withBpt(self, bpt: str) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withBpt(std::string const &) --> struct cppbinding::TwoBitToFaOption &
        """
    def withEnd(self, end: int) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withEnd(int) --> struct cppbinding::TwoBitToFaOption &
        """
    def withNoMask(self, noMask: bool) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withNoMask(bool) --> struct cppbinding::TwoBitToFaOption &
        """
    def withSeq(self, seq: str) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withSeq(std::string const &) --> struct cppbinding::TwoBitToFaOption &
        """
    def withSeqList(self, seqList: str) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withSeqList(std::string const &) --> struct cppbinding::TwoBitToFaOption &
        """
    def withStart(self, start: int) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withStart(int) --> struct cppbinding::TwoBitToFaOption &
        """
    def withUdcDir(self, udcDir: str) -> TwoBitToFaOption:
        """
        C++: cppbinding::TwoBitToFaOption::withUdcDir(std::string const &) --> struct cppbinding::TwoBitToFaOption &
        """
    @property
    def bed(self) -> str:
        """
        :type: str
        """
    @bed.setter
    def bed(self, arg0: str) -> None:
        pass
    @property
    def bedPos(self) -> bool:
        """
        :type: bool
        """
    @bedPos.setter
    def bedPos(self, arg0: bool) -> None:
        pass
    @property
    def bpt(self) -> str:
        """
        :type: str
        """
    @bpt.setter
    def bpt(self, arg0: str) -> None:
        pass
    @property
    def end(self) -> int:
        """
        :type: int
        """
    @end.setter
    def end(self, arg0: int) -> None:
        pass
    @property
    def noMask(self) -> bool:
        """
        :type: bool
        """
    @noMask.setter
    def noMask(self, arg0: bool) -> None:
        pass
    @property
    def seq(self) -> str:
        """
        :type: str
        """
    @seq.setter
    def seq(self, arg0: str) -> None:
        pass
    @property
    def seqList(self) -> str:
        """
        :type: str
        """
    @seqList.setter
    def seqList(self, arg0: str) -> None:
        pass
    @property
    def start(self) -> int:
        """
        :type: int
        """
    @start.setter
    def start(self, arg0: int) -> None:
        pass
    @property
    def udcDir(self) -> str:
        """
        :type: str
        """
    @udcDir.setter
    def udcDir(self, arg0: str) -> None:
        pass
    pass

class UsageStats:
    def __getstate__(self) -> tuple: ...
    def __init__(self) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def __str__(self) -> str: ...
    @property
    def aaCount(self) -> int:
        """
        :type: int
        """
    @aaCount.setter
    def aaCount(self, arg0: int) -> None:
        pass
    @property
    def baseCount(self) -> int:
        """
        :type: int
        """
    @baseCount.setter
    def baseCount(self, arg0: int) -> None:
        pass
    @property
    def blatCount(self) -> int:
        """
        :type: int
        """
    @blatCount.setter
    def blatCount(self, arg0: int) -> None:
        pass
    @property
    def missCount(self) -> int:
        """
        :type: int
        """
    @missCount.setter
    def missCount(self, arg0: int) -> None:
        pass
    @property
    def noSigCount(self) -> int:
        """
        :type: int
        """
    @noSigCount.setter
    def noSigCount(self, arg0: int) -> None:
        pass
    @property
    def pcrCount(self) -> int:
        """
        :type: int
        """
    @pcrCount.setter
    def pcrCount(self, arg0: int) -> None:
        pass
    @property
    def trimCount(self) -> int:
        """
        :type: int
        """
    @trimCount.setter
    def trimCount(self, arg0: int) -> None:
        pass
    @property
    def warnCount(self) -> int:
        """
        :type: int
        """
    @warnCount.setter
    def warnCount(self, arg0: int) -> None:
        pass
    pass

class ClientOption:
    def __getstate__(self) -> tuple: ...
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, arg0: ClientOption) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def __str__(self) -> str: ...
    def build(self) -> ClientOption:
        """
        C++: cppbinding::ClientOption::build() --> struct cppbinding::ClientOption &
        """
    def to_string(self) -> str:
        """
        C++: cppbinding::ClientOption::to_string() const --> std::string
        """
    def withDots(self, dots_: int) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withDots(int) --> struct cppbinding::ClientOption &
        """
    def withGenome(self, genome_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withGenome(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withGenomeDataDir(self, genomeDataDir_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withGenomeDataDir(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withHost(self, hostName_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withHost(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withInName(self, inName_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withInName(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withInSeq(self, inseq_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withInSeq(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withIsDynamic(self, isDynamic_: bool) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withIsDynamic(bool) --> struct cppbinding::ClientOption &
        """
    def withMaxIntron(self, maxIntron_: int) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withMaxIntron(long) --> struct cppbinding::ClientOption &
        """
    def withMinIdentity(self, minIdentity_: float) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withMinIdentity(double) --> struct cppbinding::ClientOption &
        """
    def withMinScore(self, minScore_: int) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withMinScore(int) --> struct cppbinding::ClientOption &
        """
    def withNohead(self, nohead_: bool) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withNohead(bool) --> struct cppbinding::ClientOption &
        """
    def withOutName(self, outName_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withOutName(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withOutputFormat(self, outputFormat_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withOutputFormat(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withPort(self, portName_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withPort(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withQType(self, qType_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withQType(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withSeqDir(self, SeqDir_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withSeqDir(const std::string &) --> struct cppbinding::ClientOption &
        """
    def withTType(self, tType_: str) -> ClientOption:
        """
        C++: cppbinding::ClientOption::withTType(const std::string &) --> struct cppbinding::ClientOption &
        """
    @property
    def SeqDir(self) -> str:
        """
        :type: str
        """
    @SeqDir.setter
    def SeqDir(self, arg0: str) -> None:
        pass
    @property
    def dots(self) -> int:
        """
        :type: int
        """
    @dots.setter
    def dots(self, arg0: int) -> None:
        pass
    @property
    def genome(self) -> str:
        """
        :type: str
        """
    @genome.setter
    def genome(self, arg0: str) -> None:
        pass
    @property
    def genomeDataDir(self) -> str:
        """
        :type: str
        """
    @genomeDataDir.setter
    def genomeDataDir(self, arg0: str) -> None:
        pass
    @property
    def hostName(self) -> str:
        """
        :type: str
        """
    @hostName.setter
    def hostName(self, arg0: str) -> None:
        pass
    @property
    def inName(self) -> str:
        """
        :type: str
        """
    @inName.setter
    def inName(self, arg0: str) -> None:
        pass
    @property
    def inSeq(self) -> str:
        """
        :type: str
        """
    @inSeq.setter
    def inSeq(self, arg0: str) -> None:
        pass
    @property
    def isDynamic(self) -> bool:
        """
        :type: bool
        """
    @isDynamic.setter
    def isDynamic(self, arg0: bool) -> None:
        pass
    @property
    def maxIntron(self) -> int:
        """
        :type: int
        """
    @maxIntron.setter
    def maxIntron(self, arg0: int) -> None:
        pass
    @property
    def minIdentity(self) -> float:
        """
        :type: float
        """
    @minIdentity.setter
    def minIdentity(self, arg0: float) -> None:
        pass
    @property
    def minScore(self) -> int:
        """
        :type: int
        """
    @minScore.setter
    def minScore(self, arg0: int) -> None:
        pass
    @property
    def nohead(self) -> bool:
        """
        :type: bool
        """
    @nohead.setter
    def nohead(self, arg0: bool) -> None:
        pass
    @property
    def outName(self) -> str:
        """
        :type: str
        """
    @outName.setter
    def outName(self, arg0: str) -> None:
        pass
    @property
    def outputFormat(self) -> str:
        """
        :type: str
        """
    @outputFormat.setter
    def outputFormat(self, arg0: str) -> None:
        pass
    @property
    def portName(self) -> str:
        """
        :type: str
        """
    @portName.setter
    def portName(self, arg0: str) -> None:
        pass
    @property
    def qType(self) -> str:
        """
        :type: str
        """
    @qType.setter
    def qType(self, arg0: str) -> None:
        pass
    @property
    def tType(self) -> str:
        """
        :type: str
        """
    @tType.setter
    def tType(self, arg0: str) -> None:
        pass
    pass

class ServerOption:
    def __getstate__(self) -> tuple: ...
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, arg0: ServerOption) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def __str__(self) -> str: ...
    def build(self) -> ServerOption:
        """
        C++: cppbinding::ServerOption::build() --> struct cppbinding::ServerOption &
        """
    def to_string(self) -> str:
        """
        C++: cppbinding::ServerOption::to_string() const --> std::string
        """
    def withCanStop(self, canStop_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withCanStop(bool) --> struct cppbinding::ServerOption &
        """
    def withDebugLog(self, debugLog_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withDebugLog(bool) --> struct cppbinding::ServerOption &
        """
    def withIndexFile(self, indexFile_: str) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withIndexFile(std::string) --> struct cppbinding::ServerOption &
        """
    def withIpLog(self, ipLog_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withIpLog(bool) --> struct cppbinding::ServerOption &
        """
    def withLog(self, log_: str) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withLog(std::string) --> struct cppbinding::ServerOption &
        """
    def withLogFacility(self, logFacility_: str) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withLogFacility(std::string) --> struct cppbinding::ServerOption &
        """
    def withMask(self, mask_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMask(bool) --> struct cppbinding::ServerOption &
        """
    def withMaxAaSize(self, maxAaSize_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMaxAaSize(int) --> struct cppbinding::ServerOption &
        """
    def withMaxDnaHits(self, maxDnaHits_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMaxDnaHits(int) --> struct cppbinding::ServerOption &
        """
    def withMaxGap(self, maxGap_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMaxGap(int) --> struct cppbinding::ServerOption &
        """
    def withMaxNtSize(self, maxNtSize_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMaxNtSize(int) --> struct cppbinding::ServerOption &
        """
    def withMaxTransHits(self, maxTransHits_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMaxTransHits(int) --> struct cppbinding::ServerOption &
        """
    def withMinMatch(self, minMatch_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withMinMatch(int) --> struct cppbinding::ServerOption &
        """
    def withNoSimpRepMask(self, noSimpRepMask_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withNoSimpRepMask(bool) --> struct cppbinding::ServerOption &
        """
    def withPerSeqMax(self, perSeqMax_: str) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withPerSeqMax(std::string) --> struct cppbinding::ServerOption &
        """
    def withRepMatch(self, repMatch_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withRepMatch(int) --> struct cppbinding::ServerOption &
        """
    def withSeqLog(self, seqLog_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withSeqLog(bool) --> struct cppbinding::ServerOption &
        """
    def withStepSize(self, stepSize_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withStepSize(int) --> struct cppbinding::ServerOption &
        """
    def withSyslog(self, syslog_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withSyslog(bool) --> struct cppbinding::ServerOption &
        """
    def withThreads(self, threads_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withThreads(int) --> struct cppbinding::ServerOption &
        """
    def withTileSize(self, tileSize_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withTileSize(int) --> struct cppbinding::ServerOption &
        """
    def withTimeout(self, timeout_: int) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withTimeout(int) --> struct cppbinding::ServerOption &
        """
    def withTrans(self, trans_: bool) -> ServerOption:
        """
        C++: cppbinding::ServerOption::withTrans(bool) --> struct cppbinding::ServerOption &
        """
    @property
    def allowOneMismatch(self) -> bool:
        """
        :type: bool
        """
    @allowOneMismatch.setter
    def allowOneMismatch(self, arg0: bool) -> None:
        pass
    @property
    def canStop(self) -> bool:
        """
        :type: bool
        """
    @canStop.setter
    def canStop(self, arg0: bool) -> None:
        pass
    @property
    def debugLog(self) -> bool:
        """
        :type: bool
        """
    @debugLog.setter
    def debugLog(self, arg0: bool) -> None:
        pass
    @property
    def genome(self) -> str:
        """
        :type: str
        """
    @genome.setter
    def genome(self, arg0: str) -> None:
        pass
    @property
    def genomeDataDir(self) -> str:
        """
        :type: str
        """
    @genomeDataDir.setter
    def genomeDataDir(self, arg0: str) -> None:
        pass
    @property
    def indexFile(self) -> str:
        """
        :type: str
        """
    @indexFile.setter
    def indexFile(self, arg0: str) -> None:
        pass
    @property
    def ipLog(self) -> bool:
        """
        :type: bool
        """
    @ipLog.setter
    def ipLog(self, arg0: bool) -> None:
        pass
    @property
    def log(self) -> str:
        """
        :type: str
        """
    @log.setter
    def log(self, arg0: str) -> None:
        pass
    @property
    def logFacility(self) -> str:
        """
        :type: str
        """
    @logFacility.setter
    def logFacility(self, arg0: str) -> None:
        pass
    @property
    def mask(self) -> bool:
        """
        :type: bool
        """
    @mask.setter
    def mask(self, arg0: bool) -> None:
        pass
    @property
    def maxAaSize(self) -> int:
        """
        :type: int
        """
    @maxAaSize.setter
    def maxAaSize(self, arg0: int) -> None:
        pass
    @property
    def maxDnaHits(self) -> int:
        """
        :type: int
        """
    @maxDnaHits.setter
    def maxDnaHits(self, arg0: int) -> None:
        pass
    @property
    def maxGap(self) -> int:
        """
        :type: int
        """
    @maxGap.setter
    def maxGap(self, arg0: int) -> None:
        pass
    @property
    def maxNtSize(self) -> int:
        """
        :type: int
        """
    @maxNtSize.setter
    def maxNtSize(self, arg0: int) -> None:
        pass
    @property
    def maxTransHits(self) -> int:
        """
        :type: int
        """
    @maxTransHits.setter
    def maxTransHits(self, arg0: int) -> None:
        pass
    @property
    def minMatch(self) -> int:
        """
        :type: int
        """
    @minMatch.setter
    def minMatch(self, arg0: int) -> None:
        pass
    @property
    def noSimpRepMask(self) -> bool:
        """
        :type: bool
        """
    @noSimpRepMask.setter
    def noSimpRepMask(self, arg0: bool) -> None:
        pass
    @property
    def perSeqMax(self) -> str:
        """
        :type: str
        """
    @perSeqMax.setter
    def perSeqMax(self, arg0: str) -> None:
        pass
    @property
    def repMatch(self) -> int:
        """
        :type: int
        """
    @repMatch.setter
    def repMatch(self, arg0: int) -> None:
        pass
    @property
    def seqLog(self) -> bool:
        """
        :type: bool
        """
    @seqLog.setter
    def seqLog(self, arg0: bool) -> None:
        pass
    @property
    def stepSize(self) -> int:
        """
        :type: int
        """
    @stepSize.setter
    def stepSize(self, arg0: int) -> None:
        pass
    @property
    def syslog(self) -> bool:
        """
        :type: bool
        """
    @syslog.setter
    def syslog(self, arg0: bool) -> None:
        pass
    @property
    def threads(self) -> int:
        """
        :type: int
        """
    @threads.setter
    def threads(self, arg0: int) -> None:
        pass
    @property
    def tileSize(self) -> int:
        """
        :type: int
        """
    @tileSize.setter
    def tileSize(self, arg0: int) -> None:
        pass
    @property
    def timeout(self) -> int:
        """
        :type: int
        """
    @timeout.setter
    def timeout(self, arg0: int) -> None:
        pass
    @property
    def trans(self) -> bool:
        """
        :type: bool
        """
    @trans.setter
    def trans(self, arg0: bool) -> None:
        pass
    pass

def buildIndex(
    gfxFile: str, fileCount: int, seqFiles: typing.List[str], options: ServerOption
) -> None:
    """
    C++: cppbinding::buildIndex(std::string &, int, class std::vector<std::string >, const struct cppbinding::ServerOption &) --> void
    """

@typing.overload
def faToTwoBit(inFiles: typing.List[str], outFile: str) -> int:
    """
    C++: cppbinding::faToTwoBit(class std::vector<std::string > &, std::string &, bool, bool, bool, bool) --> int
    """

@typing.overload
def faToTwoBit(inFiles: typing.List[str], outFile: str, noMask: bool) -> int:
    pass

@typing.overload
def faToTwoBit(
    inFiles: typing.List[str], outFile: str, noMask: bool, stripVersion: bool
) -> int:
    pass

@typing.overload
def faToTwoBit(
    inFiles: typing.List[str],
    outFile: str,
    noMask: bool,
    stripVersion: bool,
    ignoreDups: bool,
) -> int:
    pass

@typing.overload
def faToTwoBit(
    inFiles: typing.List[str],
    outFile: str,
    noMask: bool,
    stripVersion: bool,
    ignoreDups: bool,
    useLong: bool,
) -> int:
    pass

def genoFindDirect(
    probeName: str, fileCount: int, seqFiles: typing.List[str], options: ServerOption
) -> None:
    """
    C++: cppbinding::genoFindDirect(std::string &, int, class std::vector<std::string > &, const struct cppbinding::ServerOption &) --> void
    """

def genoPcrDirect(
    fPrimer: str,
    rPrimer: str,
    fileCount: int,
    seqFiles: typing.List[str],
    options: ServerOption,
) -> None:
    """
    C++: cppbinding::genoPcrDirect(std::string &, std::string &, int, class std::vector<std::string > &, const struct cppbinding::ServerOption &) --> void
    """

def getFileList(hostName: str, portName: str) -> None:
    """
    C++: cppbinding::getFileList(std::string &, std::string &) --> void
    """

def getPortIx(portName: str) -> int:
    """
    C++: cppbinding::getPortIx(char *) --> int
    """

def gfServer(options: ServerOption) -> None:
    """
    C++: cppbinding::gfServer(struct cppbinding::ServerOption &) --> void
    """

def pcrServer(
    hostName: str, portName: str, fPrimer: str, rPrimer: str, maxSize: int
) -> None:
    """
    C++: cppbinding::pcrServer(std::string &, std::string &, std::string &, std::string &, int) --> void
    """

def pygetFileList(hostName: str, portName: str) -> str:
    """
    C++: cppbinding::pygetFileList(std::string &, std::string &) --> std::string
    """

def pygfClient(option: ClientOption) -> bytes:
    """
    C++: cppbinding::pygfClient(struct cppbinding::ClientOption &) --> std::string
    """

def pygfClient2(arg0: ClientOption) -> str:
    pass

def pygfClient_no_gil(option: ClientOption) -> None:
    """
    C++: cppbinding::pygfClient(struct cppbinding::ClientOption &) --> std::string
    """

def pyqueryServer(
    type: str, hostName: str, portName: str, faName: str, complex: bool, isProt: bool
) -> str:
    """
    C++: cppbinding::pyqueryServer(std::string &, std::string &, std::string &, std::string &, bool, bool) --> std::string
    """

def pystartServer(
    hostName: str,
    portName: str,
    fileCount: int,
    seqFiles: typing.List[str],
    options: ServerOption,
    stats: UsageStats,
) -> int:
    """
    C++: cppbinding::pystartServer(std::string &, std::string &, int, class std::vector<std::string > &, struct cppbinding::ServerOption &, struct cppbinding::UsageStats &) --> int
    """

def pystartServer_no_gil(
    hostName: str,
    portName: str,
    fileCount: int,
    seqFiles: typing.List[str],
    options: ServerOption,
    stats: UsageStats,
) -> int:
    """
    C++: cppbinding::pystartServer(std::string &, std::string &, int, class std::vector<std::string > &, struct cppbinding::ServerOption &, struct cppbinding::UsageStats &) --> int
    """

def pystatusServer(hostName: str, portName: str, options: ServerOption) -> str:
    """
    C++: cppbinding::pystatusServer(std::string &, std::string &, struct cppbinding::ServerOption &) --> std::string
    """

def queryServer(
    type: str, hostName: str, portName: str, faName: str, complex: bool, isProt: bool
) -> None:
    """
    C++: cppbinding::queryServer(std::string &, std::string &, std::string &, std::string &, bool, bool) --> void
    """

def startServer(
    hostName: str,
    portName: str,
    fileCount: int,
    seqFiles: typing.List[str],
    options: ServerOption,
    stats: UsageStats,
) -> None:
    """
    C++: cppbinding::startServer(std::string &, std::string &, int, class std::vector<std::string > &, struct cppbinding::ServerOption &, struct cppbinding::UsageStats &) --> void
    """

def statusServer(hostName: str, portName: str, options: ServerOption) -> int:
    """
    C++: cppbinding::statusServer(std::string &, std::string &, struct cppbinding::ServerOption &) --> int
    """

def stopServer(hostName: str, portName: str) -> None:
    """
    C++: cppbinding::stopServer(std::string &, std::string &) --> void
    """

def test(arg0: int, arg1: IntStruct) -> None:
    pass

def test_add(a: int) -> None:
    """
    C++: cppbinding::test_add(int &) --> void
    """

def test_exception() -> None:
    """
    C++: cppbinding::test_exception() --> void
    """

def test_no_gil(arg0: int, arg1: IntStruct) -> None:
    pass

def test_stat(stats: UsageStats) -> None:
    """
    C++: cppbinding::test_stat(struct cppbinding::UsageStats &) --> void
    """

def test_stdout() -> None:
    """
    C++: cppbinding::test_stdout() --> void
    """

def test_with_gil(arg0: int, arg1: IntStruct) -> None:
    pass

def twoBitToFa(inName: str, outName: str, option: TwoBitToFaOption) -> None:
    """
    C++ cppbinding::twoBitToFa(std::string cppinName, std::string cppoutName, TwoBitToFaOption option)
    """



"""Command line interface for pxblat."""
import sys
from datetime import datetime

import typer

from pxblat import __version__

from .client import client
from .fa2twobit import faToTwoBit
from .server import server_app
from .twobit2fa import twoBitToFa

app = typer.Typer(
    epilog=f"YangyangLi {datetime.now().year} yangyang.li@northwstern.edu",
    context_settings={"help_option_names": ["-h", "--help"]},
)

app.command()(faToTwoBit)
app.command()(client)
app.command()(twoBitToFa)


app.add_typer(server_app, name="server")


@app.callback()
def common():  # noqa: D103
    pass


common.__doc__ = f"""{typer.style("Version", fg=typer.colors.YELLOW, bold=True)}: {typer.style(f"{__version__}", fg=typer.colors.GREEN, bold=True)}"""


if "sphinx" in sys.modules and __name__ != "__main__":
    # Create the typer click object to generate docs with sphinx-click
    typer_click_object = typer.main.get_command(app)


if __name__ == "__main__":
    app()



from pathlib import Path

import typer

from pxblat import fa_to_two_bit

from .log import logger

# faToTwoBit - Convert DNA from fasta to 2bit format
# usage:
#    faToTwoBit in.fa [in2.fa in3.fa ...] out.2bit
# options:
#    -long          use 64-bit offsets for index.   Allow for twoBit to contain more than 4Gb of sequence.
#                   NOT COMPATIBLE WITH OLDER CODE.
#    -noMask        Ignore lower-case masking in fa file.
#    -stripVersion  Strip off version number after '.' for GenBank accessions.
#    -ignoreDups    Convert first sequence only if there are duplicate sequence
#                   names.  Use 'twoBitDup' to find duplicate sequences.


def faToTwoBit(
    infa: list[Path] = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        readable=True,
        metavar="in.fa [inf2.fa in3.fa ...]",
        show_default=False,
        help="The fasta files",
    ),
    out2bit: Path = typer.Argument(
        ...,
        dir_okay=False,
        writable=True,
        show_default=False,
        metavar="out.2bit",
        help="The output file",
    ),
    long: bool = typer.Option(
        False,
        "--long",
        help="Use 64-bit offsets for index. Allow for twoBit to contain more than 4Gb of sequence.",
    ),
    noMask: bool = typer.Option(
        False,
        "--nomask",
        help="Ignore lower-case masking in fa file.",
    ),
    stripVersion: bool = typer.Option(
        False,
        "--stripVersion",
        help="Strip off version number after '.' for GenBank accessions.",
    ),
    ignoreDups: bool = typer.Option(
        False,
        "--ignoreDups",
        help="Convert first sequence only if there are duplicate sequence names. Use 'twoBitDup' to find duplicate sequences.",
    ),
):
    """Convert DNA from fasta to 2bit format."""
    for file in infa:
        if not file.exists():
            logger.error(f"{file} not exists")
            raise typer.Abort()
        if not file.is_file():
            logger.error(f"{file} is not a file")
            raise typer.Abort()

    if out2bit.exists():
        logger.warning(f"{out2bit} exist will be override")

    fa_to_two_bit(
        [f.as_posix() for f in infa],
        out2bit.as_posix(),
        noMask=noMask,
        stripVersion=stripVersion,
        ignoreDups=ignoreDups,
        useLong=long,
    )



from pathlib import Path

import typer
from rich import print

from pxblat.extc import UsageStats
from pxblat.server import (
    create_server_option,
    start_server_mt,
    status_server,
    stop_server,
)
from pxblat.server import files as server_files

# gfServer v 37x1 - Make a server to quickly find where DNA occurs in genome

#    To set up a server:
#       gfServer start host port file(s)
#       where the files are .2bit or .nib format files specified relative to the current directory

#    To remove a server:
#       gfServer stop host port

#    To query a server with DNA sequence:
#       gfServer query host port probe.fa

#    To query a server with protein sequence:
#       gfServer protQuery host port probe.fa

#    To query a server with translated DNA sequence:
#       gfServer transQuery host port probe.fa

#    To query server with PCR primers:
#       gfServer pcr host port fPrimer rPrimer maxDistance

#    To process one probe fa file against a .2bit format genome (not starting server):
#       gfServer direct probe.fa file(s).2bit

#    To test PCR without starting server:
#       gfServer pcrDirect fPrimer rPrimer file(s).2bit

#    To figure out if server is alive, on static instances get usage statics as well:
#       gfServer status host port
#      For dynamic gfServer instances, specify -genome and optionally the -genomeDataDir
#      to get information on an untranslated genome index. Include -trans to get about information
#      about a translated genome index

#    To get input file list:
#       gfServer files host port

#    To generate a precomputed index:
#       gfServer index gfidx file(s)
#      where the files are .2bit or .nib format files.  Separate indexes are
#      be created for untranslated and translated queries.  These can be used
#      with a persistent server as with 'start -indexFile or a dynamic server.
#      They must follow the naming convention for for dynamic servers.

#    To run a dynamic server (usually called by xinetd):
#       gfServer dynserver rootdir
#      Data files for genomes are found relative to the root directory.
#      Queries are made using the prefix of the file path relative to the root
#      directory.  The files $genome.2bit, $genome.untrans.gfidx, and
#      $genome.trans.gfidx are required. Typically the structure will be in
#      the form:
#          $rootdir/$genomeDataDir/$genome.2bit
#          $rootdir/$genomeDataDir/$genome.untrans.gfidx
#          $rootdir/$genomeDataDir/$genome.trans.gfidx
#      in this case, one would call gfClient with
#          -genome=$genome -genomeDataDir=$genomeDataDir
#      Often $genomeDataDir will be the same name as $genome, however it
#      can be a multi-level path. For instance:
#           GCA/902/686/455/GCA_902686455.1_mSciVul1.1/
#      The translated or untranslated index maybe omitted if there is no
#      need to handle that type of request.
#      The -perSeqMax functionality can be implemented by creating a file
#          $rootdir/$genomeDataDir/$genome.perseqmax


default_option = create_server_option()

server_app = typer.Typer(
    help="Make a server to quickly find where DNA occurs in genome",
)


tileSize: int = typer.Option(
    default_option.tileSize,
    "--tile-size",
    help="Size of n-mers to index.  Default is 11 for nucleotides, 4 for proteins (or translated nucleotides).",
)

stepSize: int = typer.Option(
    default_option.stepSize,
    "--stepSize",
    help="Spacing between tiles. Default is tileSize.",
)

minMatch: int = typer.Option(
    default_option.minMatch,
    "--min-match",
    help="Number of n-mer matches that trigger detailed alignment. Default is 2 for nucleotides, 3 for proteins.",
)

trans: bool = typer.Option(
    default_option.trans,
    "--trans",
    help="Translate database to protein in 6 frames.",
)

log: str = typer.Option(
    None,
    "--log",
    exists=True,
    dir_okay=False,
    help="Keep a log file that records server requests.",
)

mask: bool = typer.Option(
    default_option.mask,
    "--mask",
    help="Use masking from .2bit file.",
)

repMatch: int = typer.Option(
    default_option.repMatch,
    "--repMatch",
    help="Number of occurrences of a tile (n-mer) that triggers repeat masking the tile. Default is 1024.",
)

noSimpRepMask: bool = typer.Option(
    default_option.noSimpRepMask,
    "--noSimpRepMask",
    help="Suppresses simple repeat masking.",
)

maxDnaHits: int = typer.Option(
    default_option.maxDnaHits,
    "--maxDnaHits",
    help="Maximum number of hits for a DNA query that are sent from the server.",
)

maxTransHits: int = typer.Option(
    default_option.maxTransHits,
    "--maxTransHits",
    help="Maximum number of hits for a translated query that are sent from the server.",
)

maxNtSize: int = typer.Option(
    default_option.maxNtSize,
    "--maxNtSize",
    help="Maximum size of untranslated DNA query sequence.",
)

perSeqMax: Path = typer.Option(
    None,
    "--perSeqMax",
    exists=True,
    dir_okay=False,
    help="File contains one seq filename (possibly with ':seq' suffix) per line.",
)

canStop: bool = typer.Option(
    default_option.canStop,
    "--canStop",
    help="If set, a quit message will actually take down the server.",
)

indexFile: Path = typer.Option(
    None,
    "--indexFile",
    exists=True,
    dir_okay=False,
    help="Index file create by `gfServer index'.",
)
timeout: int = typer.Option(
    default_option.timeout,
    "--timeout",
    help="Timeout in seconds.",
)


two_bit: Path = typer.Argument(
    ...,
    exists=True,
    dir_okay=False,
    help="Two bit file",
)

threads: int = typer.Option(2, "--threads", help="Number of threads to use")


@server_app.command()
def start(
    host: str,
    port: int,
    two_bit: Path = two_bit,
    tileSize: int = tileSize,
    stepSize: int = stepSize,
    minMatch: int = minMatch,
    trans: bool = trans,
    log: str = log,
    mask: bool = mask,
    repMatch: int = repMatch,
    noSimpRepMask: bool = noSimpRepMask,
    maxDnaHits: int = maxDnaHits,
    maxTransHits: int = maxTransHits,
    maxNtSize: int = maxNtSize,
    perSeqMax: Path = perSeqMax,
    canStop: bool = canStop,
    indexFile: Path = indexFile,
    timeout: int = timeout,
):
    """To set up a server.

    gfServer start host port file(s)

    where the files are .2bit or .nib format files specified relative to the current directory
    """
    server_option = (
        create_server_option()
        .withTileSize(tileSize)
        .withStepSize(stepSize)
        .withMinMatch(minMatch)
        .withTrans(trans)
        .withMask(mask)
        .withRepMatch(repMatch)
        .withNoSimpRepMask(noSimpRepMask)
        .withMaxDnaHits(maxDnaHits)
        .withMaxTransHits(maxTransHits)
        .withMaxNtSize(maxNtSize)
        .withCanStop(canStop)
        .withTimeout(timeout)
    )

    if log is not None:
        server_option.withLog(log)

    if perSeqMax is not None:
        server_option.withPerSeqMax(perSeqMax.as_posix())

    if indexFile is not None:
        server_option.withIndexFile(indexFile.as_posix())

    server_option.build()

    stat = UsageStats()

    start_server_mt(
        host,
        port,
        two_bit.as_posix(),
        server_option,
        stat,
        try_new_port=False,
    )


@server_app.command()
def stop(host: str, port: int):
    """To remove a server."""
    stop_server(host, port)


@server_app.command()
def status(
    host: str,
    port: int,
    trans: bool = trans,
):
    """To figure out if server is alive, on static instances get usage statics."""
    server_option = create_server_option().withTrans(trans)
    server_option.build()
    ret = status_server(host, port, server_option)
    print(ret)


@server_app.command()
def files(host: str, port: int):
    """To get input file list."""
    ret = server_files(host, port)
    print(ret)



from .._extc.cppbinding import *  # type: ignore


def set_state(self, d):
    self.hostName = d[0]
    self.portName = d[1]
    self.tType = d[2]
    self.qType = d[3]
    self.dots = d[4]
    self.nohead = d[5]
    self.minScore = d[6]
    self.minIdentity = d[7]
    self.outputFormat = d[8]
    self.maxIntron = d[9]
    self.genome = d[10]
    self.genomeDataDir = d[11]
    self.isDynamic = d[12]
    self.SeqDir = d[13]
    self.inName = d[14]
    self.outName = d[15]
    self.inSeq = d[16]


ClientOption.__setstate__ = set_state  # type: ignore



from __future__ import annotations

import tempfile
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING, Union

from gevent.pool import Pool

from pxblat.extc import ClientOption, pygfClient
from pxblat.parser import read

from .basic import wait_server_ready

if TYPE_CHECKING:
    from .server import ServerOption

INSEQ = Union[str, Path]
INSEQS = Union[list[INSEQ], list[str], list[Path]]


def copy_client_option(option: ClientOption) -> ClientOption:
    """Copies the ClientOption object."""
    new_option = ClientOption()
    new_option.hostName = option.hostName
    new_option.portName = option.portName
    new_option.tType = option.tType
    new_option.qType = option.qType
    new_option.dots = option.dots
    new_option.nohead = option.nohead
    new_option.minScore = option.minScore
    new_option.minIdentity = option.minIdentity
    new_option.outputFormat = option.outputFormat
    new_option.maxIntron = option.maxIntron
    new_option.genome = option.genome
    new_option.genomeDataDir = option.genomeDataDir
    new_option.isDynamic = option.isDynamic
    new_option.SeqDir = option.SeqDir
    new_option.inName = option.inName
    new_option.outName = option.outName
    new_option.inSeq = option.inSeq
    return new_option


def create_client_option():
    """Creates a new ClientOption object with default values.

    Return:
        ClientOption object

    See Also:
        :class:`.ClientOption`

    Examples:
        >>> option = create_client_option().build()
        >>> option
        ClientOption(hostName=, portName=, tType=dna, qType=dna, dots=0, nohead=false, minScore=30, minIdentity=90,
                     outputFormat=psl, maxIntron=750000, genome=, genomeDataDir=, isDynamic=false,
                     tSeqDir=, inName=, outName=)
        >>> option = create_client_option().withPort("66666").build()
        >>> option
        ClientOption(hostName=, portName=66666, tType=dna, qType=dna, dots=0, nohead=false, minScore=30, minIdentity=90,
                     outputFormat=psl, maxIntron=750000, genome=, genomeDataDir=, isDynamic=false,
                     tSeqDir=, inName=, outName=)
    """
    return ClientOption()


def _resolve_host_port(
    client_option: ClientOption,
    host: str | None,
    port: int | None,
):
    """Resolves the host and port for the client option.

    Args:
        client_option: ClientOption
        host: Optional[str]
        port: Optional[int]
    """
    if host is not None:
        client_option.hostName = host

    if port is not None:
        client_option.portName = str(port)

    if not client_option.hostName and not client_option.portName:
        msg = "host and port are both empty"
        raise ValueError(msg)


def query_server_by_file(
    option: ClientOption,
    host: str | None = None,
    port: int | None = None,
    *,
    parse: bool = True,
):
    """Sends a query to the server and returns the result.

    Args:
        option: ClientOption
        host: Optional[str]
        port: Optional[int]
        seqname: Optional[str]
        parse: bool

    Returns:
        str or bytes: The result of the query.

    """
    _resolve_host_port(option, host, port)

    ret = pygfClient(option)

    try:
        ret_decode = ret.decode().rsplit(",\n", 1)[0]  # type: ignore
    except UnicodeDecodeError:
        ret_decode = ret.decode("latin-1").rsplit(",\n", 1)[0]  # type: ignore

    if parse and ret_decode:
        try:
            ret = read(ret_decode, "psl")
        except ValueError as e:
            if "No query results" in str(e):
                return None
        else:
            return ret

    return ret_decode


def _assign_info_to_query_result(query_result):
    query_result.version = "v.37x1"
    return query_result


def query_server(
    option: ClientOption,
    host: str | None = None,
    port: int | None = None,
    seqname: str | None = None,
    *,
    parse: bool = True,
):
    """Sends a query to the server and returns the result.

    Args:
        option: ClientOption
        host: Optional[str]
        port: Optional[int]
        seqname: Optional[str]
        parse: bool

    Returns:
        str or bytes: The result of the query.

    """
    _resolve_host_port(option, host, port)

    fafile = None
    if not option.inName and not option.inSeq:
        msg = "inName and inSeq are both empty"
        raise ValueError(msg)

    if option.inSeq:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as fafile:
            seqname = fafile.name if seqname is None else seqname
            fafile.write(f">{seqname}\n")
            fafile.write(option.inSeq)
        option.inName = fafile.name

    ret = pygfClient(option)

    try:
        ret_decode = ret.decode().rsplit(",\n", 1)[0]  # type: ignore
    except UnicodeDecodeError:
        ret_decode = ret.decode("latin-1").rsplit(",\n", 1)[0]  # type: ignore

    if fafile is not None:
        Path(fafile.name).unlink()

    if parse and ret_decode:
        try:
            res = read(ret_decode, "psl")
        except ValueError as e:
            if "No query results" in str(e):
                return None
        else:
            if isinstance(res, str):
                return _assign_info_to_query_result(read(res, "psl"))
            return _assign_info_to_query_result(res)

    return ret_decode


class ClientThread(Thread):
    """A class for managing client connections to a server in a separate thread.

    This class can be used to query a gfServer in a separate thread, and can optionally wait until the server is ready before
    sending a query. It can also parse the result of the query.

    Attributes:
        option (ClientOption): Client options for the connection.
        host (str, optional): The hostname or IP address of the server. Defaults to None.
        port (int, optional): The port number of the server. Defaults to None.
        wait_ready (bool, optional): Whether to wait until the server is ready before sending a query. Defaults to False.
        wait_timeout (int, optional): The number of seconds to wait for the server to be ready. Defaults to 60.
        server_option (ServerOption, optional): The server options to use if a server is not provided. Defaults to None.
        seqname (str, optional): The sequence name to use for the query. Defaults to None.
        parse (bool, optional): Whether to parse the result of the query. Defaults to True.
        daemon (bool, optional): Whether to run the client as a daemon process. Defaults to True.
        result: The result of the query, or None if the query has not yet been sent or the result has not yet been received.

    Order:
        -10
    """

    def __init__(
        self,
        option: ClientOption,
        host: str | None = None,
        port: int | None = None,
        *,
        wait_timeout: int = 60,
        server_option: ServerOption | None = None,
        seqname: str | None = None,
        wait_ready: bool = False,
        parse: bool = True,
        daemon: bool = True,
    ) -> None:
        """A class for querying a gfServer using a separate thread.

        Args:
            option: ClientOption
            host: Optional[str]
            port: Optional[int]
            wait_ready: bool
            wait_timeout: int
            server_option: Optional[ServerOption]
            seqname: Optional[str]
            parse: bool
            daemon: bool

        Attributes:
            result: The result of the query.

        """
        super().__init__(daemon=daemon)
        self.option = option
        self._host = host
        self._port = port
        self._resolve_host_port()

        self._wait_ready = wait_ready
        self._wait_timeout = wait_timeout
        self._server_option = server_option
        self._seqname = seqname
        self._parse = parse

        self.result = None

    def run(self):
        """Runs the query in a separate thread."""
        if self._wait_ready:
            wait_server_ready(
                self.host,
                self.port,
                timeout=self._wait_timeout,
                server_option=self._server_option,
            )

        ret = query_server(self.option, seqname=self._seqname, parse=self._parse)

        self.result = ret

    def get(self):
        """Sends a query to the server and returns the result."""
        self.join()
        return self.result

    @property
    def host(self):
        """The hostname or IP address of the server."""
        if self._host is None:
            return self.option.hostName

        return self._host

    @host.setter
    def host(self, value: str):
        """Sets the hostname or IP address of the server."""
        self._host = value

    @property
    def port(self):
        """The port number of the server."""
        if self._port is None:
            return int(self.option.portName)

        return self._port

    @port.setter
    def port(self, value: int):
        """Sets the port number of the server."""
        self._port = value

    @classmethod
    def create_option(cls):
        """Creates a new ClientOption object with default values.

        Return:
            ClientOption object

        """
        return create_client_option()

    def _resolve_host_port(self):
        _resolve_host_port(self.option, self._host, self._port)


class Client:
    """A class for managing client connections to a server in a separate thread.

    This class can be used to query a gfServer in a separate thread, and can optionally wait until the server is ready before
    sending a query. It can also parse the result of the query.

    Attributes:
        option (ClientOption): Client options for the connection.
        host (str, optional): The hostname or IP address of the server. Defaults to None.
        port (int, optional): The port number of the server. Defaults to None.
        wait_ready (bool, optional): Whether to wait until the server is ready before sending a query. Defaults to False.
        wait_timeout (int, optional): The number of seconds to wait for the server to be ready. Defaults to 60.
        server_option (ServerOption, optional): The server options to use if a server is not provided. Defaults to None.
        seqname (str, optional): The sequence name to use for the query. Defaults to None.
        parse (bool, optional): Whether to parse the result of the query. Defaults to True.
        daemon (bool, optional): Whether to run the client as a daemon process. Defaults to True.
        result: The result of the query, or None if the query has not yet been sent or the result has not yet been received.

    Order:
        -10
    """

    def __init__(
        self,
        host: str,
        port: int,
        seq_dir: str | Path,
        *,
        ttype: str = "dna",
        qtype: str = "dna",
        dots: int = 0,
        nohead: bool = False,
        min_score: int = 30,
        min_identity: float = 90.0,
        output_format: str = "psl",
        max_intron: int = 750000,
        is_dynamic: bool = False,
        genome: str | None = None,
        genome_data_dir: str | None = None,
        server_option: ServerOption | None = None,
        wait_ready: bool = False,
        wait_timeout: int = 60,
        parse: bool = True,
    ) -> None:
        """A class for querying a gfServer using a separate thread.

        Args:
            host (str): The hostname or IP address of the server.
            port (int): The port number of the server.
            seq_dir (Union[str, Path]): The directory where sequence data is stored.
            ttype (str, optional): Database type. One of 'dna', 'prot', 'dnax'. Default is 'dna'.
            qtype (str, optional): Query type. One of 'dna', 'rna', 'prot', 'dnax', 'rnax'. Default is 'dna'.
            dots (int, optional): Output a dot every N query sequences. Default is 0.
            nohead (bool, optional): If True, suppresses 5-line psl header. Default is False.
            min_score (int, optional): Sets minimum score. Default is 30.
            min_identity (float, optional): Sets minimum sequence identity (in percent). Default is 90.
            output_format (str, optional): Controls output file format. One of 'psl', 'pslx', 'axt', 'maf', 'sim4', 'wublast', 'blast', 'blast8', 'blast9'. Default is 'psl'.
            max_intron (int, optional): Sets maximum intron size. Default is 750000.
            is_dynamic (bool, optional): If True, the client is expected to interact with a dynamic gfServer. Default is False.
            genome (Optional[str], optional): The genome name when using a dynamic gfServer. Defaults to None.
            genome_data_dir (Optional[str], optional): The root directory containing the genome data files for a dynamic gfServer. Defaults to None.
            server_option (Optional[ServerOption], optional): The server options to use if a server is not provided. Defaults to None.
            wait_ready (bool, optional): If True, wait until the server is ready before sending a query. Default is False.
            wait_timeout (int, optional): The number of seconds to wait for the server to be ready. Default is 60.
            parse (bool, optional): If True, parse the result of the query. Default is True.

        Raises:
            ValueError: If any of the input values are invalid.

        Examples:
            >>> from pxblat import Client
            >>> host = "localhost"
            >>> port = 65000
            >>> seq_dir = "."
            >>> two_bit = "./test_ref.2bit"
            >>> client = Client(
            ...     host=host,
            ...     port=port,
            ...     seq_dir=seq_dir,
            ...     min_score=20,
            ...     min_identity=90,
            ... )
        """
        self._basic_option = (
            ClientOption()
            .withHost(host)
            .withPort(str(port))
            .withMinScore(min_score)
            .withMinIdentity(min_identity)
            .withTType(ttype)
            .withQType(qtype)
            .withDots(dots)
            .withNohead(nohead)
            .withMaxIntron(max_intron)
            .withOutputFormat(output_format)
            .withIsDynamic(is_dynamic)
        )

        if genome is not None:
            self._basic_option.withGenome(genome)
        if genome_data_dir is not None:
            self._basic_option.withGenomeDataDir(genome_data_dir)

        self._basic_option.withSeqDir(str(seq_dir))

        self._wait_ready = wait_ready
        self._wait_timeout = wait_timeout
        self._server_option = server_option
        self._parse = parse

    # fmt: off
    @property
    def seq_dir(self):
        """The directory containing the sequence files."""
        return self._basic_option.SeqDir
    @seq_dir.setter
    def seq_dir(self, value: str | Path): self._basic_option.withSeqDir(str(value))

    @property
    def ttype(self):
        """The type of the target sequence."""
        return self._basic_option.tType
    @ttype.setter
    def ttype(self, value: str): self._basic_option.withTType(value)

    @property
    def qtype(self):
        """The type of the query sequence."""
        return self._basic_option.qType
    @qtype.setter
    def qtype(self, value: str): self._basic_option.withQType(value)

    @property
    def min_score(self):
        """The minimum score for the alignment."""
        return int(self._basic_option.minScore)
    @min_score.setter
    def min_score(self, value: int): self._basic_option.withMinScore(value)

    @property
    def min_identity(self):
        """The minimum identity for the alignment."""
        return self._basic_option.minIdentity
    @min_identity.setter
    def min_identity(self, value: float): self._basic_option.withMinIdentity(value)

    @property
    def host(self):
        """The hostname or IP address of the server."""
        return self._basic_option.hostName
    @host.setter
    def host(self, value: str): self._basic_option.withHost(value)

    @property
    def port(self):
        """The port number of the server."""
        return int(self._basic_option.portName)
    @port.setter
    def port(self, value: int): self._basic_option.withPort(str(value))

    @property
    def output_format(self):
        """The output format of the alignment."""
        return self._basic_option.outputFormat
    @output_format.setter
    def output_format(self, value: str): self._basic_option.withOutputFormat(value)

    @property
    def max_intron(self):
        """The maximum intron size for the alignment."""
        return self._basic_option.maxIntron
    @max_intron.setter
    def max_intron(self, value: int): self._basic_option.withMaxIntron(value)

    @property
    def is_dynamic(self):
        """Whether the server is dynamic."""
        return self._basic_option.isDynamic
    @is_dynamic.setter
    def is_dynamic(self, value: bool): self._basic_option.withIsDynamic(value)

    @property
    def genome(self):
        """The genome name of the server."""
        return self._basic_option.genome
    @genome.setter
    def genome(self, value: str): self._basic_option.withGenome(value)

    @property
    def genome_data_dir(self):
        """The genome data directory of the server."""
        return self._basic_option.genomeDataDir
    @genome_data_dir.setter
    def genome_data_dir(self, value: str): self._basic_option.withGenomeDataDir(value)
    # fmt: on

    @staticmethod
    def _verify_input(in_seqs: list[str | Path] | list[str] | list[Path]):
        for item in in_seqs:
            if isinstance(item, Path) and not item.exists():
                msg = f"File {item} does not exist"
                raise FileNotFoundError(msg)

            if isinstance(item, str) and ("." in item or "/" in item):
                new_item = Path(item)
                if not new_item.exists():
                    msg = f"File {item} does not exist"
                    raise FileNotFoundError(msg)

                yield new_item

            yield item

    def _query(self, in_seq: str | Path):
        basic_option = copy_client_option(self._basic_option)
        if isinstance(in_seq, Path):
            basic_option.withInName(str(in_seq)).withInSeq("").build()
        else:
            basic_option.withInSeq(str(in_seq)).withInName("").build()
        return query_server(basic_option, parse=self._parse)

    def query(self, in_seqs: INSEQS | list[str] | list[Path] | INSEQ):
        """Query the server with the specified sequences.

        Args:
            in_seqs: The sequences to query.

        Returns:
            The query results: `Bio.SearchIO.QueryResult`

        Examples:
            >>> from pxblat import Client, Server
            >>> host = "localhost"
            >>> port = 65000
            >>> seq_dir = "."
            >>> two_bit = "./test_ref.2bit"
            >>> client = Client(
            ...     host=host,
            ...     port=port,
            ...     seq_dir=seq_dir,
            ...     min_score=20,
            ...     min_identity=90,
            ... )
            >>> with Server(host, port, two_bit, can_stop=True, step_size=5) as server:
            ...     # work() assume work() is your own function that takes time to prepare something
            ...     server.wait_ready()
            ...     result1 = client.query("ATCG")
            ...     result2 = client.query("AtcG")
            ...     result3 = client.query("test_case1.fa")
            ...     result4 = client.query(["ATCG", "ATCG"])
            ...     result5 = client.query(["test_case1.fa"])
            ...     result6 = client.query(["cgTA", "test_case1.fa"])
            ...     print(result3[0]) # print result
        """
        if isinstance(in_seqs, (str, Path)):
            in_seqs = [in_seqs]

        in_seqs = list(self._verify_input(in_seqs))

        if self._wait_ready:
            wait_server_ready(
                self.host,
                self.port,
                timeout=self._wait_timeout,
                server_option=self._server_option,
            )

        group = Pool(1)
        return list(group.imap(self._query, in_seqs))



from pathlib import Path

import typer

from pxblat.extc import pygfClient
from pxblat.server import create_client_option

# gfClient v. 37x1 - A client for the genomic finding program that produces a .psl file
# usage:
#    gfClient host port seqDir in.fa out.psl
# where
#    host is the name of the machine running the gfServer
#    port is the same port that you started the gfServer with
#    seqDir is the path of the .2bit or .nib files relative to the current dir
#        (note these are needed by the client as well as the server)
#    in.fa is a fasta format file.  May contain multiple records
#    out.psl is where to put the output
# options:
#    -t=type       Database type. Type is one of:
#                    dna - DNA sequence
#                    prot - protein sequence
#                    dnax - DNA sequence translated in six frames to protein
#                  The default is dna.
#    -q=type       Query type. Type is one of:
#                    dna - DNA sequence
#                    rna - RNA sequence
#                    prot - protein sequence
#                    dnax - DNA sequence translated in six frames to protein
#                    rnax - DNA sequence translated in three frames to protein
#    -prot         Synonymous with -t=prot -q=prot.
#    -dots=N       Output a dot every N query sequences.
#    -nohead       Suppresses 5-line psl header.
#    -minScore=N   Sets minimum score.  This is twice the matches minus the
#                  mismatches minus some sort of gap penalty.  Default is 30.
#    -minIdentity=N   Sets minimum sequence identity (in percent).  Default is
#                  90 for nucleotide searches, 25 for protein or translated
#                  protein searches.
#    -out=type     Controls output file format.  Type is one of:
#                    psl - Default.  Tab-separated format without actual sequence
#                    pslx - Tab-separated format with sequence
#                    axt - blastz-associated axt format
#                    maf - multiz-associated maf format
#                    sim4 - similar to sim4 format
#                    wublast - similar to wublast format
#                    blast - similar to NCBI blast format
#                    blast8- NCBI blast tabular format
#                    blast9 - NCBI blast tabular format with comments
#    -maxIntron=N   Sets maximum intron size. Default is 750000.
#    -genome=name  When using a dynamic gfServer, The genome name is used to
#                  find the data files relative to the dynamic gfServer root, named
#                  in the form $genome.2bit, $genome.untrans.gfidx, and $genome.trans.gfidx
#                  When using a dynamic gfServer, this is the dynamic gfServer root directory
#                  that contained the genome data files.  Defaults to being the root directory.


default_option = create_client_option()


def client(
    host: str = typer.Argument(
        ...,
        help="The name of the machine running the gfServer",
    ),
    port: int = typer.Argument(
        ...,
        help="The same port that you started the gfServer with",
    ),
    seqdir: Path = typer.Argument(
        ...,
        dir_okay=True,
        help="The path of the .2bit or .nib files relative to the current dir",
    ),
    infasta: Path = typer.Argument(
        ...,
        help="Fasta format file.  May contain multiple records",
    ),
    outpsl: Path = typer.Argument(..., help="where to put the output"),
    tType: str = typer.Option(
        default_option.tType,
        "--type",
        "-t",
        help="Database type. Type is one of: dna, prot, dnax",
    ),
    qType: str = typer.Option(
        default_option.qType,
        "--qtype",
        "-q",
        help="Query type. Type is one of: dna, rna, prot, dnax, rnax",
    ),
    prot: bool = typer.Option(False, "--prot", help="Synonymous with -t=prot -q=prot."),
    dots: int = typer.Option(
        default_option.dots,
        "--dots",
        help="Output a dot every N query sequences.",
    ),
    nohead: bool = typer.Option(
        default_option.nohead,
        "--nohead",
        help="Suppresses 5-line psl header.",
    ),
    minnScore: int = typer.Option(
        default_option.minScore,
        "--minScore",
        help="Sets minimum score.  This is twice the matches minus the mismatches minus some sort of gap penalty.  Default is 30.",
    ),
    minIdentity: int = typer.Option(
        default_option.minIdentity,
        "--minIdentity",
        help="Sets minimum sequence identity (in percent).  Default is 90 for nucleotide searches, 25 for protein or translated protein searches.",
    ),
    out: str = typer.Option(
        default_option.outputFormat,
        "--out",
        help="Controls output file format.  Type is one of: psl, pslx, axt, maf, sim4, wublast, blast, blast8, blast9",
    ),
    maxIntron: int = typer.Option(
        default_option.maxIntron,
        "--maxIntron",
        help="Sets maximum intron size. Default is 750000.",
    ),
    genome: str = typer.Option(default_option.genome, "--genome", help="dynamic"),
    genomeDataDir: str = typer.Option(
        default_option.genomeDataDir,
        "--genomeDataDir",
        help="dynamic",
    ),
):
    """A client for the genomic finding program that produces a .psl file."""
    if prot:
        tType = "prot"
        qType = "prot"

    client_option = (
        create_client_option()
        .withHost(host)
        .withPort(str(port))
        .withSeqDir(seqdir.as_posix())
        .withInName(infasta.as_posix())
        .withOutName(outpsl.as_posix())
        .withTType(tType)
        .withQType(qType)
        .withDots(dots)
        .withNohead(nohead)
        .withMinScore(minnScore)
        .withMinIdentity(minIdentity)
        .withOutputFormat(out)
        .withMaxIntron(maxIntron)
        .withGenome(genome)
        .withGenomeDataDir(genomeDataDir)
        .build()
    )

    pygfClient(client_option)



from .extc import ClientOption
from .extc import ServerOption
from .extc import TwoBitToFaOption
from .extc import UsageStats
from .parser import read
from .server import check_port_in_use
from .server import check_port_open
from .server import Client
from .server import ClientThread
from .server import create_client_option
from .server import create_server_option
from .server import copy_client_option
from .server import files
from .server import find_free_port
from .server import query_server
from .server import Server
from .server import server_query
from .server import start_server
from .server import start_server_mt
from .server import start_server_mt_nb
from .server import Status
from .server import status_server
from .server import stop_server
from .server import wait_server_ready
from .toolkit import fa_to_two_bit
from .toolkit import two_bit_to_fa
from rich.traceback import install

__version__ = "1.0.0"

install(show_locals=True)


__all__ = [
    "Server",
    "ClientThread",
    "Client",
    "files",
    "server_query",
    "start_server_mt",
    "start_server",
    "status_server",
    "stop_server",
    "create_client_option",
    "query_server",
    "fa_to_two_bit",
    "two_bit_to_fa",
    "create_server_option",
    "check_port_open",
    "start_server_mt_nb",
    "wait_server_ready",
    "find_free_port",
    "check_port_in_use",
    "copy_client_option",
    "Status",
    "read",
    "TwoBitToFaOption",
    "ClientOption",
    "ServerOption",
    "UsageStats",
]



from __future__ import annotations

from pathlib import Path

from Bio.SearchIO.BlatIO import BlatPslParser  # type: ignore

# https://github.com/biopython/biopython/blob/master/Bio/SearchIO/BlatIO.py


class PslOutput:
    """A class for PSL output."""

    def __init__(self, content: str, *, save_mem: bool = False) -> None:
        """Initialize the class."""
        self._content: str | None = None if save_mem else content
        self._iterable = iter(content.splitlines(keepends=True))

    def readline(self):
        """Read a line from the file."""
        try:
            return next(self._iterable)
        except StopIteration:
            return ""


_ITERATOR_MAP = {
    "psl": BlatPslParser,
}

_HANDLE_MAP = {
    "psl": PslOutput,
}


def get_handle(format: str, mapping):
    """Returns the handle associated with the given format from the mapping.

    This function takes a format string and a mapping from format strings to handles, and returns the handle associated
    with the given format. If the format is not in the mapping, it raises an error with a helpful message.

    Args:
        format (str): The format string for which to get the handle. This should be a lower case string.
        mapping (dict): The mapping from format strings to handles.

    Raises:
        ValueError: If the format is None, not a lower case string, or not in the mapping.
        TypeError: If the format is not a string.

    Returns:
        The handle associated with the given format.
    """
    try:
        handle = mapping[format]
    except KeyError:
        # handle the errors with helpful messages
        if format is None:
            raise ValueError("Format required (lower case string)") from None

        if not isinstance(format, str):
            raise TypeError("Need a string for the file format (lower case)") from None
        if format != format.lower():
            raise ValueError("Format string %r should be lower case" % format) from None

        msg = f"Unknown format {format!r}. Supported formats are {', '.join(mapping)}"
        raise ValueError(msg) from None

    else:
        return handle


def get_processor(format, mapping):
    """Returns the information object associated with the given format from the mapping.

    This function takes a format string and a mapping from format strings to information objects,
    and returns the information object associated with the given format. If the format is not in the mapping,
    it raises an error with a helpful message.

    Args:
        format (str): The format string for which to get the information object. This should be a lower case string.
        mapping (dict): The mapping from format strings to information objects.

    Raises:
        ValueError: If the format is None, not a lower case string, or not in the mapping.
        TypeError: If the format is not a string.

    Returns:
        obj_info: The information object associated with the given format.
    """
    try:
        obj_info = mapping[format]
    except KeyError:
        # handle the errors with helpful messages
        if format is None:
            msg = "Format required (lower case string)"
            raise ValueError(msg) from None
        if not isinstance(format, str):
            msg = "Need a string for the file format (lower case)"
            raise TypeError(msg) from None
        if format != format.lower():
            raise ValueError("Format string %r should be lower case" % format) from None

        msg = f"Unknown format {format!r}. Supported formats are {', '.join(mapping)}"
        raise ValueError(msg) from None

    else:
        return obj_info


def parse(content: str, format=None, **kwargs):
    """Parses the given content according to the specified format.

    This function takes a string content and a format string, gets the corresponding iterator and handle using the
    format from the '_ITERATOR_MAP' and '_HANDLE_MAP' respectively, and then yields the parsed content using the
    iterator.

    Args:
        content (str): The string content to parse.
        format (str, optional): The format string indicating how to parse the content. If not provided, 'psl' format will be used. Defaults to None.
        **kwargs: Arbitrary keyword arguments to be passed to the iterator function.

    Yields:
        The parsed content.

    Raises:
        ValueError: If the format is None after defaulting, not a lower case string, or not in the mapping.
        TypeError: If the format is not a string.
    """
    if format is None:
        format = "psl"

    iterator = get_processor(format, _ITERATOR_MAP)
    handle = get_handle(format, _HANDLE_MAP)
    yield from iterator(handle(content), **kwargs)


def read(content: str, format=None, **kwargs):
    """Reads and returns the first query result from the given content.

    This function takes a string content and a format string, parses the content using the `parse` function,
    and returns the first query result. If no results are found, or if more than one result is found, it raises an error.

    Args:
        content (str): The string content to parse and read.
        format (str, optional): The format string indicating how to parse the content. If not provided, 'psl' format will be used. Defaults to None.
        **kwargs: Arbitrary keyword arguments to be passed to the `parse` function.

    Returns:
        query_result: The first query result found in the content.

    Raises:
        ValueError: If no query results are found in the content, or if more than one query result is found.
        ValueError/TypeError: If there is an error in parsing the content. These exceptions are propagated from the `parse` function.
    """
    query_results = parse(content, format, **kwargs)
    try:
        query_result = next(query_results)
    except StopIteration:
        msg = "No query results found in handle"
        raise ValueError(msg) from None

    try:
        next(query_results)
        msg = "More than one query result found in handle"
        raise ValueError(msg)
    except StopIteration:
        pass

    return query_result


def _psl2sam(psl: str, samfile: Path):
    raise NotImplementedError


def psl2sam(psl: str | Path, samfile: Path):
    """Converts a psl file to a sam file."""
    if isinstance(psl, Path):
        psl = psl.read_text()
    elif Path(psl).exists():
        psl = Path(psl).read_text()

    _psl2sam(psl, samfile)



"""Server module."""
from .basic import (
    check_port_in_use,
    check_port_open,
    files,
    find_free_port,
    server_query,
    start_server,
    start_server_mt,
    start_server_mt_nb,
    status_server,
    stop_server,
    wait_server_ready,
)
from .client import (
    Client,
    ClientThread,
    copy_client_option,
    create_client_option,
    query_server,
)
from .server import Server, create_server_option
from .status import Status

__all__ = [
    "Server",
    "ClientThread",
    "files",
    "server_query",
    "start_server_mt",
    "start_server",
    "status_server",
    "stop_server",
    "create_client_option",
    "query_server",
    "create_server_option",
    "check_port_open",
    "start_server_mt_nb",
    "wait_server_ready",
    "find_free_port",
    "check_port_in_use",
    "copy_client_option",
    "Status",
    "Client",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pxblat', 'pxblat.cli', 'pxblat.extc', 'pxblat.server', 'pxblat.toolkit']

package_data = \
{'': ['*'],
 'pxblat.extc': ['bindings/*',
                 'bindings/binder/*',
                 'include/*',
                 'include/aux/*',
                 'include/core/*',
                 'include/net/*',
                 'src/*',
                 'src/aux/*',
                 'src/core/*',
                 'src/net/*']}

install_requires = \
['biopython>=1.81,<2.0',
 'deprecated>=1.2.13,<2.0.0',
 'gevent==22.10.2',
 'loguru>=0.7.0,<0.8.0',
 'mashumaro>=3.7,<4.0',
 'pybind11>=2.10.4,<3.0.0',
 'pysimdjson>=5.0.2,<6.0.0',
 'rich>=13.3.5,<14.0.0',
 'setuptools>=68.2.2,<69.0.0',
 'typer>=0.9.0,<0.10.0',
 'urllib3==2.0.7']

entry_points = \
{'console_scripts': ['pxblat = pxblat.cli.cli:app']}

setup_kwargs = {
    'name': 'pxblat',
    'version': '1.0.0',
    'description': 'A native python binding for blat suit',
    'long_description': '# <img src="https://raw.githubusercontent.com/cauliyang/pxblat/main/docs/_static/logo.png" alt="logo" height=100> **PxBLAT** [![social](https://img.shields.io/github/stars/cauliyang/pxblat?style=social)](https://github.com/cauliyang/pxblat/stargazers)\n\n_An Efficient and Ergonomic Python Binding Library for BLAT_\n\n[![python](https://img.shields.io/badge/Python-3776AB.svg?style=for-the-badge&logo=Python&logoColor=white)](https://www.python.org/)\n[![c++](https://img.shields.io/badge/C++-00599C.svg?style=for-the-badge&logo=C++&logoColor=white)](https://en.cppreference.com/w/)\n[![c](https://img.shields.io/badge/C-A8B9CC.svg?style=for-the-badge&logo=C&logoColor=black)](https://www.gnu.org/software/gnu-c-manual/)\n[![pypi](https://img.shields.io/pypi/v/pxblat.svg?style=for-the-badge)][pypi]\n[![conda](https://img.shields.io/conda/vn/bioconda/pxblat?style=for-the-badge)][conda]\n![Linux](https://img.shields.io/badge/-Linux-grey?logo=linux&style=for-the-badge)\n![macOS](https://img.shields.io/badge/-OSX-black?logo=apple&style=for-the-badge)\n[![pyversion](https://img.shields.io/pypi/pyversions/pxblat?style=for-the-badge)][pypi]\n[![tests](https://img.shields.io/github/actions/workflow/status/cauliyang/pxblat/tests.yml?style=for-the-badge&logo=github&label=Tests)](https://github.com/cauliyang/pxblat/actions/workflows/tests.yml)\n[![Codecov](https://img.shields.io/codecov/c/github/cauliyang/pxblat/main?style=for-the-badge)](https://app.codecov.io/gh/cauliyang/pxblat)\n[![docs](https://img.shields.io/readthedocs/pxblat?style=for-the-badge)](https://pxblat.readthedocs.io/en/latest/)\n[![download](https://img.shields.io/pypi/dm/pxblat?logo=pypi&label=pypi%20download&style=for-the-badge)][pypi]\n[![condadownload](https://img.shields.io/conda/dn/bioconda/pxblat?style=for-the-badge&logo=anaconda&label=Conda%20Download)][conda]\n[![precommit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?style=for-the-badge&logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![black](https://img.shields.io/badge/code%20style-ruff-000000.svg?style=for-the-badge)](https://github.com/charliermarsh/ruff)\n[![Ruff](https://img.shields.io/endpoint?url=https://raw.githubusercontent.com/charliermarsh/ruff/main/assets/badge/v1.json&style=for-the-badge)](https://github.com/charliermarsh/ruff)\n[![release](https://img.shields.io/github/release-date/cauliyang/pxblat?style=for-the-badge)](https://github.com/cauliyang/pxblat/releases)\n[![open-issue](https://img.shields.io/github/issues-raw/cauliyang/pxblat?style=for-the-badge)][open-issue]\n[![close-issue](https://img.shields.io/github/issues-closed-raw/cauliyang/pxblat?style=for-the-badge)][close-issue]\n[![activity](https://img.shields.io/github/commit-activity/m/cauliyang/pxblat?style=for-the-badge)][repo]\n[![lastcommit](https://img.shields.io/github/last-commit/cauliyang/pxblat?style=for-the-badge)][repo]\n[![opull](https://img.shields.io/github/issues-pr-raw/cauliyang/pxblat?style=for-the-badge)][opull]\n[![all contributors](https://img.shields.io/github/all-contributors/cauliyang/pxblat?style=for-the-badge)](#contributors)\n\n<!-- [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)][colab] -->\n\n[repo]: https://github.com/ylab-hi/pxblat\n[open-issue]: https://github.com/cauliyang/pxblat/issues?q=is%3Aissue+is%3Aopen+sort%3Aupdated-desc\n[close-issue]: https://github.com/cauliyang/pxblat/issues?q=is%3Aissue+sort%3Aupdated-desc+is%3Aclosed\n[opull]: https://github.com/cauliyang/pxblat/pulls?q=is%3Apr+is%3Aopen+sort%3Aupdated-desc\n[conda]: https://bioconda.github.io/recipes/pxblat/README.html\n[pypi]: https://pypi.org/project/pxblat/\n[colab]: https://colab.research.google.com/drive/1TXb9GBmYa2EYezwBKbD-y9Xg6MC2gL36\n\n## Why PxBLAT?\n\nWhen conducting extensive queries, using the `blat` of `BLAT` suit can prove to be quite inefficient, especially if these operations aren\'t grouped. The tasks are allocated sporadically, often interspersed among other tasks.\nIn general, the choice narrows down to either utilizing `blat` or combining `gfServer` with `gfClient`.\nIndeed, `blat` is a program that launches `gfServer`, conducts the sequence query via `gfClient`, and then proceeds to terminate the server.\n\nThis approach is far from ideal when performing numerous queries that aren\'t grouped since `blat` repeatedly initializes and shuts down `gfServer` for each query, resulting in substantial overhead.\nThis overhead consists of the time required for the server to index the reference, contingent on the reference\'s size.\nTo index the human genome (hg38), for example, would take approximately five minutes.\n\nA more efficient solution would involve initializing `gfServer` once and invoking `gfClient` multiple times for the queries.\nHowever, `gfServer` and `gfClient` are only accessible via the command line.\nThis necessitates managing system calls (for instance, `subprocess` or `os.system`), intermediate temporary files, and format conversion, further diminishing performance.\n\nThat is why `PxBLAT` holds its position.\nIt resolves the issues mentioned above while introducing handy features like `port retry`, `use current running server`, etc.\n\n## 📚 **Table of Contents**\n\n- [ **PxBLAT** ](#-pxblat-)\n  - [📚 **Table of Contents**](#-table-of-contents)\n  - [🔮 **Features**](#-features)\n  - [📎 **Citation**](#-citation)\n  - [🚀 **Getting Started**](#-getting-started)\n  - [🤝 **Contributing**](#-contributing)\n  - [\U0001faaa **License**](#-license)\n  - [**Contributors**](#contributors)\n  - [🙏 **Acknowledgments**](#-acknowledgments)\n\n## 🔮 **Features**\n\n- **Zero System Calls**: Avoids system calls, leading to a smoother, quicker operation.<br>\n- **Ergonomics**: With an ergonomic design, `PxBLAT` aims for a seamless user experience.<br>\n- **No External Dependencies**: `PxBLAT` operates independently without any external dependencies.<br>\n- **Self-Monitoring**: No need to trawl through log files; `PxBLAT` monitors its status internally.<br>\n- **Robust Validation**: Extensively tested to ensure reliable performance and superior stability as BLAT.<br>\n- **Format-Agnostic:** `PxBLAT` doesn\'t require you to worry about file formats.<br>\n- **In-Memory Processing**: `PxBLAT` discards the need for intermediate files by doing all its operations in memory, ensuring speed and efficiency.<br>\n\n## 📎 **Citation**\n\nPxBLAT is scientific software, with a published paper in the BioRxiv.\nCheck the [published](https://www.biorxiv.org/content/10.1101/2023.08.02.551686v2) to read the paper.\n\n```bibtex\n@article {Li2023pxblat,\n\tauthor = {Yangyang Li and Rendong Yang},\n\ttitle = {PxBLAT: An Ergonomic and Efficient Python Binding Library for BLAT},\n\telocation-id = {2023.08.02.551686},\n\tyear = {2023},\n\tdoi = {10.1101/2023.08.02.551686},\n\tpublisher = {Cold Spring Harbor Laboratory},\n\turl = {https://www.biorxiv.org/content/10.1101/2023.08.02.551686v2},\n\tjournal = {bioRxiv}\n}\n```\n\n## 🚀 **Getting Started**\n\nPlease see the [document](https://pxblat.readthedocs.io/en/latest/) for details and more examples.\n\n## 🤝 **Contributing**\n\nContributions are always welcome! Please follow these steps:\n\n1. Fork the project repository. This creates a copy of the project on your account that you can modify without affecting the original project.\n2. Clone the forked repository to your local machine using a Git client like Git or GitHub Desktop.\n3. Create a new branch with a descriptive name (e.g., `new-feature-branch` or `bugfix-issue-123`).\n\n```bash\ngit checkout -b new-feature-branch\n```\n\n4. Take changes to the project\'s codebase.\n5. Install the latest package\n\n```bash\npoetry install\n```\n\n6. Test your changes\n\n```bash\npytest -vlsx tests\n```\n\n7. Commit your changes to your local branch with a clear commit message that explains the changes you\'ve made.\n\n```bash\ngit commit -m \'Implemented new feature.\'\n```\n\n8. Push your changes to your forked repository on GitHub using the following command\n\n```bash\ngit push origin new-feature-branch\n```\n\nCreate a pull request to the original repository.\nOpen a new pull request to the original project repository. In the pull request, describe the changes you\'ve made and why they\'re necessary.\nThe project maintainers will review your changes and provide feedback or merge them into the main branch.\n\n## \U0001faaa **License**\n\n**PxBLAT is modified from blat, the license is the same as blat. The source code and\nexecutables are freely available for academic, nonprofit, and personal use.\nCommercial licensing information is available on the Kent Informatics website\n(https://kentinformatics.com/).**\n\n## **Contributors**\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tbody>\n    <tr>\n      <td align="center" valign="top" width="14.28%"><a href="https://yangyangli.top"><img src="https://avatars.githubusercontent.com/u/38903141?v=4?s=100" width="100px;" alt="yangliz5"/><br /><sub><b>yangliz5</b></sub></a><br /><a href="#maintenance-cauliyang" title="Maintenance">🚧</a></td>\n      <td align="center" valign="top" width="14.28%"><a href="https://github.com/mencian"><img src="https://avatars.githubusercontent.com/u/71105179?v=4?s=100" width="100px;" alt="Joshua Zhuang"/><br /><sub><b>Joshua Zhuang</b></sub></a><br /><a href="#infra-mencian" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n    </tr>\n  </tbody>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\n## 🙏 **Acknowledgments**\n\n- [BLAT](http://genome.ucsc.edu/goldenPath/help/blatSpec.html)\n- [UCSC](https://github.com/ucscGenomeBrowser/kent)\n- [pybind11](https://github.com/pybind/pybind11/tree/stable)\n\n<!-- github-only -->\n\n<br>\n<picture>\n  <source media="(prefers-color-scheme: dark)" srcset="https://api.star-history.com/svg?repos=cauliyang/pxblat&type=Date&theme=light" />\n  <source media="(prefers-color-scheme: light)" srcset="https://api.star-history.com/svg?repos=cauliyang/pxblat&type=Date" />\n  <img alt="Star History Chart" src="https://api.star-history.com/svg?repos=cauliyang/pxblat&type=Date" />\n</picture>\n',
    'author': 'Yangyang Li',
    'author_email': 'yangyang.li@northwestern.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ylab-hi/pxblat',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.12',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from __future__ import annotations

import errno
import socket
import time
import warnings
from collections import Counter
from multiprocessing import Process

from deprecated import deprecated  # type: ignore

from pxblat.extc import (
    ServerOption,
    UsageStats,
    pygetFileList,
    pyqueryServer,
    pystartServer,
    startServer,
)

from .status import Status

MAX_PORT = 65535


def check_port_in_use(host: str, port: int, tries: int = 3) -> bool:
    """Check if a given port on a host is in use.

    Args:
        host (str): The hostname or IP address of the host to check.
        port (int, optional): The port number to check. Defaults to DEFAULT_PORT.
        tries (int, optional): The number of times to attempt the check. Defaults to 3.

    Returns:
        bool: True if the port is in use, False otherwise.

    Raises:
        TypeError: If the host argument is not a string.
        ValueError: If the port argument is not a valid port number.

    This function attempts to connect to the specified host and port using the socket library. It will attempt the connection
    a number of times specified by the tries argument. If the connection is successful, it will return True, indicating that
    the port is in use. If the connection fails, it will return False.

    If the host argument is not a string, a TypeError will be raised. If the port argument is not a valid port number, a
    ValueError will be raised.

    Example:
        >>> check_port_in_use('localhost', 8080)
        True
    """
    res = [_check_port_in_use_by_connect(host, port) for _ in range(tries)]
    counter = Counter(res)
    return counter[True] > counter[False]


def _check_port_in_use_by_status(
    host: str,
    port: int,
    server_option: ServerOption,
) -> bool:
    """Check the port is in use by status_server.

    Args:
        host: host name
        port: port number
        server_option: server option for the opening server

    Returns:
        True if the port is in use

    Note:
        The server option can be default value and will not influence result
        Also, it can detect if the port is opened by gfServer as well
    """
    return check_server_status(host, port, server_option)


def _check_port_in_use_by_connect(host: str, port: int):
    """Check the port is in use by connect to the port.

    Args:
        host: host name
        port: port number

    Returns:
        True if the port is in use

    Note:
        The function has same feature as `_check_port_in_use_by_status`
        It check the port if it is in use by connect to the port.
        But it cannot detect if the port is opened by gfServer
    """
    return check_port_open(host, port)


@deprecated(
    reason="The func will generate false alarm. Please use `_check_port_in_use_by_status` or  `_check_port_in_use_by_connect` instead",
)
def _check_port_in_use_by_bind(host: str, port: int):
    """Check the port is in use by bind to the port.

    Args:
        host: host name
        port: port number

    Returns:
        True if the port is in use

    Note:
        The function check the port if it is in use by bind to the port.
        It is not reliable as ``_check_port_in_use_by_connect`` or
        ``_check_port_in_use_by_status``

    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind((host, port))
    except OSError as e:
        if e.errno == errno.EADDRINUSE:
            return True

        raise e

    s.close()
    return False


def wait_server_ready(
    host: str,
    port: int,
    timeout: int = 60,
    server_option=None,
) -> None:
    """Wait for a server to become ready by checking if a given port is open or if a specific server status is reached.

    Args:
        host (str): The hostname or IP address of the server to check.
        port (int): The port number to check for open status.
        timeout (int, optional): The maximum number of seconds to wait for the server to become ready. Defaults to 60.
        server_option (str, optional): The specific server status to check for. If None, the function will check for an open port. Defaults to None.

    Raises:
        RuntimeError: If the server does not become ready within the specified timeout.

    Returns:
        None
    """
    start = time.perf_counter()

    if server_option is None:
        warnings.warn(
            "Use `check_server_status` instead of `check_port_open` when wait for ready",
            stacklevel=1,
        )
        while not check_port_open(host, port):
            time.sleep(1)
            if time.perf_counter() - start > timeout:
                msg = "wait for server ready timeout"
                raise RuntimeError(msg)
    else:
        while not check_server_status(host, port, server_option):
            time.sleep(1)
            if time.perf_counter() - start > timeout:
                msg = "wait for server ready timeout"
                raise RuntimeError(msg)


def check_server_status(
    host: str,
    port: int,
    server_option: ServerOption,
) -> bool:
    """Check the status of a server by attempting to connect to it using the specified host, port, and gfserver_option.

    Args:
        host (str): The hostname or IP address of the server to check.
        port (int): The port number to use when attempting to connect to the server.
        server_option (ServerOption): The gfserver option to use when attempting to connect to the server.

    Returns:
        bool: True if the server is running and accepting connections, False otherwise.

    Raises:
        ConnectionRefusedError: If the server is not running or is not accepting connections.

    Example:
        >>> check_server_status('localhost', 8080, ServerOption)
        True
    """
    try:
        status_server(host, port, server_option)
    except ConnectionRefusedError:
        return False
    else:
        return True


def check_port_open(host: str, port: int) -> bool:
    """Check the port is open and can accept message or not.

    Args:
        host: Hostname
        port: Port number

    Returns:
        True if the port is open and can accept message, otherwise False.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def _gfSignature() -> str:
    """Get the gfSignature."""
    return "0ddf270562684f29"


def status_server(
    host: str,
    port: int,
    server_option: ServerOption,
    *,
    instance=False,
) -> Status | dict[str, str]:
    """Get the status of a running server.

    Args:
        host (str): The hostname or IP address of the server to check.
        port (int): The port number to use when attempting to connect to the server.
        server_option (ServerOption): The gfserver option to use when attempting to connect to the server.
        instance (bool, optional): If True, return a Status object instead of a dictionary. Defaults to False.

    Returns:
        Union[Status, Dict[str, str]]: A dictionary or Status object containing the status information for the server.

    Example:
        >>> status_server('localhost', 8080, ServerOption, instance=True)
        Status(uptime='0', queries='0', sequences='0', bytes='0', memory='0', threads='0', connections='0')
    """
    if not server_option.genome:
        message = f"{_gfSignature()}status".encode()
    else:
        temp = "transInfo" if server_option.trans else "untransInfo"
        message = f"{_gfSignature()}{temp} {server_option.genome} {server_option.genomeDataDir}".encode()

    data = b""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((host, port))
        s.sendall(message)
        while True:
            data += s.recv(1024)
            if b"end" in data:
                break

        mapping = {
            b"\x00": b"\n",
            b"\x03": b"\n",
            b"\x04": b"\n",
            b"\x07": b"\n",
            b"\x08": b"\n",
            b"\x11": b"\n",
            b"\x10": b"\n",
            b"\x0b": b"\n",
            b"\x0c": b"\n",
            b"\x0e": b"\n",
            b"\x0f": b"\n",
            b"\t": b"\n",
            b"end": b"",
        }

    for k, v in mapping.items():
        data = data.replace(k, v)

    data_dict = {" ".join(line.split()[:-1]): line.split()[-1] for line in data.decode("utf-8").strip().split("\n")}

    if instance:
        return Status.from_dict(data_dict)

    return data_dict


def stop_server(host: str, port: int):
    """Stop a running server.

    Args:
        host (str): The hostname or IP address of the server to stop.
        port (int): The port number to use when attempting to connect to the server.

    Returns:
        None

    This function stops a running server by sending a "quit" message to the server. The function takes the hostname and port number
    of the server as arguments. The function returns None.

    Example:
        >>> stop_server('localhost', 8080)
    """
    message = f"{_gfSignature()}quit".encode()
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((host, port))
        s.sendall(message)


def files(host: str, port: int) -> list[str]:
    """Get a list of files available on the server.

    Args:
        host (str): The hostname or IP address of the server to check.
        port (int): The port number to use when attempting to connect to the server.

    Returns:
        List[str]: A list of file names available on the server.

    This function retrieves a list of files available on the server by sending a "files" message to the server. The function takes
    the hostname and port number of the server as arguments. The function returns a list of file names available on the server.

    Example:
        >>> files('localhost', 8080)
        ['file1', 'file2', 'file3']
    """
    ret_str = pygetFileList(host, str(port))
    if ret_str == "":
        raise ValueError
    return [file for file in ret_str.split("\n") if file]


def server_query(
    intype: str,
    host: str,
    port: int,
    faName: str,
    *,
    isComplex: bool,
    isProt: bool,
):
    """Query a running server with a sequence.

    Args:
        intype (str): The type of input sequence. Must be one of 'dna', 'rna', or 'protein'.
        host (str): The hostname or IP address of the server to query.
        port (int): The port number to use when attempting to connect to the server.
        faName (str): The name of the input sequence.
        isComplex (bool): Whether the input sequence is complex.
        isProt (bool): Whether the input sequence is a protein sequence.

    Returns:
        str: The result of the query.

    This function queries a running server with a sequence by sending a "query" message to the server. The function takes the
    type of input sequence, hostname, port number, sequence name, and whether the sequence is complex or a protein sequence as
    arguments. The function returns the result of the query as a string.

    Example:
        >>> server_query('dna', 'localhost', 8080, 'sequence1', False, False)
        'result1'
    """
    return pyqueryServer(intype, host, str(port), faName, isComplex, isProt)


def start_server(
    host: str,
    port: int,
    two_bit_file: str,
    server_option: ServerOption,
    stat: UsageStats,
):
    """Start a server in blocking mode.

    Args:
        host (str): The hostname or IP address to bind the server to.
        port (int): The port number to bind the server to.
        two_bit_file (str): The path to the 2bit file to use for the server.
        server_option (ServerOption): The options to use for the server.
        stat (UsageStats): The usage statistics for the server.

    Returns:
        None

    This function starts a server in blocking mode by calling the pystartServer function with the given arguments. The function
    takes the hostname, port number, 2bit file path, server options, and usage statistics as arguments. The function returns None.

    Example:
        >>> start_server('localhost', 8080, '/path/to/2bit/file', ServerOption, UsageStats())
    """
    return startServer(host, str(port), 1, [two_bit_file], server_option, stat)


def start_server_mt(
    host: str,
    port: int,
    two_bit_file: str,
    server_option: ServerOption,
    stat: UsageStats,
    *,
    use_others: bool = False,
    try_new_port: bool = True,
):
    """Starts a server on a new thread.

    This function attempts to start a server on the given host and port, using the provided .2bit file, options, and usage stats.
    If the port is in use, it either waits for the server to be ready, tries a new port, or raises an error, depending on the options.

    Args:
        host (str): The hostname or IP address to start the server on.
        port (int): The initial port number to try to start the server on.
        two_bit_file (str): The .2bit file to use for the server.
        server_option (ServerOption): The options to use when starting the server.
        stat (UsageStats): The usage stats to use for the server.
        use_others (bool, optional): Whether to use the port even if it is already in use. Defaults to False.
        try_new_port (bool, optional): Whether to try a new port if the initial port is in use. Defaults to True.

    Raises:
        ValueError: If the port is in use and neither 'use_others' nor 'try_new_port' is True.
        Exception: If there is any other error in starting the server.
    """
    try:
        if check_port_in_use(host, port):
            if use_others:
                pass
            elif try_new_port:
                port = find_free_port(host, start=port + 1)
                pystartServer(host, str(port), 1, [two_bit_file], server_option, stat)
            else:
                msg = f"The port {port} is used"
                raise ValueError(msg)
        else:
            pystartServer(host, str(port), 1, [two_bit_file], server_option, stat)
    except Exception as e:
        raise e


def start_server_mt_nb(
    host: str,
    port: int,
    two_bit_file: str,
    server_option: ServerOption,
    stat: UsageStats,
    *,
    use_others: bool = False,
    try_new_port: bool = True,
) -> Process:
    """Starts a server on a new thread and immediately returns a Process object.

    This function calls `start_server_mt` to start a server on the given host and port, using the provided .2bit file, options, and usage stats,
    in a new thread, and immediately returns a Process object for the new thread.

    Args:
        host (str): The hostname or IP address to start the server on.
        port (int): The initial port number to try to start the server on.
        two_bit_file (str): The .2bit file to use for the server.
        server_option (ServerOption): The options to use when starting the server.
        stat (UsageStats): The usage stats to use for the server.
        use_others (bool, optional): Whether to use the port even if it is already in use. Defaults to False.
        try_new_port (bool, optional): Whether to try a new port if the initial port is in use. Defaults to True.

    Returns:
        Process: A Process object for the newly started thread.
    """
    process = Process(
        target=start_server_mt,
        args=(
            host,
            port,
            two_bit_file,
            server_option,
            stat,
        ),
        kwargs={
            "use_others": use_others,
            "try_new_port": try_new_port,
        },
        daemon=True,
    )

    process.start()
    return process


def find_free_port(host: str, start: int, end: int = MAX_PORT) -> int:
    """Find an available port in the range of [start, end).

    Args:
        host: Hostname
        start: Start port number
        end: End port number.
    """
    if start > end:
        raise ValueError

    for port in range(start, end):
        try:
            if not check_port_in_use(host, port):
                return port
        except OSError as e:
            raise e
    msg = f"Cannot find available port in range [{start}, {end}]"
    raise RuntimeError(msg)
